(************************************************************************
 Author:	Robert Rossmair
		Wagelwitz, Rodaer Str. 6
		04688 Mutzschen
		Germany

	e-mail:	Robert.Rossmair@t-online.de
	http://home.t-online.de/home/Robert.Rossmair/

 Module:	SplitBox

 Version:	26-OCT-1997

 Description:	class TSplitBox: panel with tiling capabilities

 History:
   26-oct-97:	new property "PasteTarget", useful at design time only
   21-oct-97:	published Division property to make it persistent
   15-oct-97:	minor changes
   09-oct-97:
	unit "Tilers" renamed to "SplitBox",
	class "TTileBox" renamed to "TSplitBox" etc.
	splitter appearance changed
	detected that MinSize{LeftTop|RightBottom} had no effect;
	changed that
   01-feb-97:
	final release of unit Tilers

 Copyright  1997  Robert Rossmair

 Use, modification and distribution is allowed without limitation,
 warranty, or liability of any kind.
 If you use this software or portions of it in a commercial program,
 I would appreciate it if you mentioned my contribution in the
 documentation materials accompaning your work.
 ************************************************************************)
unit SplitBox;

interface

uses
  SysUtils,	Windows,	Messages,
  Classes,	Graphics,	Controls,
  Forms,	ExtCtrls,	Math32;

type
  TSplitType = (stHorizontal, stVertical);

  TSplitBoxPartition = class(TCustomPanel)
  end;

  TSplitBox = class;

  TPartitionIndex = 1..2;
  TPartitionVisibilityEvent = procedure(Sender: TSplitBox; Partition: TPartitionIndex) of object;
  TPasteTarget = (ptLeftTopPartition, ptRightBottomPartition);

  TSplitBox = class(TCustomControl)
  private
    FPartition: array[TPartitionIndex] of TSplitBoxPartition;
    FSplitterRect: TRect;
    FMinSize1: Integer;
    FMinSize2: Integer;
    FPartition1Count: Integer;
    FOnShowPartition: TPartitionVisibilityEvent;
    FOnHidePartition: TPartitionVisibilityEvent;
    FDivision: Integer;
    FMovePos: TPoint;
    FPasteTarget: TPasteTarget;
    FSplitType: TSplitType;
    FTransparent: Boolean;
    FSplitterDragging: Boolean;
    procedure ShowPartition(Index: TPartitionIndex);
    procedure HidePartition(Index: TPartitionIndex);
    function GetBevelInner: TPanelBevel;
    function GetBevelOuter: TPanelBevel;
    function GetBevelWidth: TBevelWidth;
    function GetBorderWidth: TBorderWidth;
    function GetDivision: Single;
    function GetMinSizeLeftTop: NaturalNumber;
    function GetMinSizeRightBottom: NaturalNumber;
//    procedure ReadDivision(Reader: TReader);
    procedure SetBevelInner(Value: TPanelBevel);
    procedure SetBevelOuter(Value: TPanelBevel);
    procedure SetBevelWidth(Value: TBevelWidth);
    procedure SetDivision(Value: Single);
    procedure SetBorderWidth(Value: TBorderWidth);
    procedure SetMinSizeLeftTop(Value: NaturalNumber);
    procedure SetMinSizeRightBottom(Value: NaturalNumber);
    procedure SetPasteTarget(Value: TPasteTarget);
    procedure SetSplitType(Value: TSplitType);
    procedure SetTransparent(Value: Boolean);
    procedure UpdatePartitions;
    procedure UpdateHorz;
    procedure UpdateVert;
    procedure UpdateDivision;
    procedure UpdateTransparent;
//    procedure WriteDivision(Writer: TWriter);
    procedure CMDesignHitTest(var Msg: TCMDesignHitTest); message CM_DESIGNHITTEST;
    procedure WMSetCursor(var Msg: TWMSetCursor); message WM_SETCURSOR;
  protected
    procedure AlignControls(AControl: TControl; var Rect: TRect); override;
    procedure CreateParams(var Params: TCreateParams); override;
    procedure GetChildren(Proc: TGetChildProc; Root: TComponent); override;
//    procedure DefineProperties(Filer: TFiler); override;
    procedure Loaded; override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure Paint; override;
    procedure ReadState(Reader: TReader); override;
    procedure WndProc(var Message: TMessage); override;
    procedure WriteState(Writer: TWriter); override;
    property Cursor default crHSplit;
  public
    constructor Create(AOwner: TComponent); override;
    procedure AddPartition1(Control: TControl);
    procedure AddPartition2(Control: TControl);
    property Partition1: TSplitBoxPartition read FPartition[1];
    property Partition2: TSplitBoxPartition read FPartition[2];
  published
    property Align;
    property BevelInner: TPanelBevel read GetBevelInner write SetBevelInner default bvLowered;
    property BevelOuter: TPanelBevel read GetBevelOuter write SetBevelOuter default bvNone;
    property BevelWidth: TBevelWidth read GetBevelWidth write SetBevelWidth default 1;
    property BorderWidth: TBorderWidth read GetBorderWidth write SetBorderWidth default 0;
    property Division: Single read GetDivision write SetDivision;
    property MinSizeLeftTop: NaturalNumber read GetMinSizeLeftTop write SetMinSizeLeftTop default 32;
    property MinSizeRightBottom: NaturalNumber read GetMinSizeRightBottom write SetMinSizeRightBottom default 32;
    // useful at design time only
    property PasteTarget: TPasteTarget read FPasteTarget write SetPasteTarget stored False;
    property SplitType: TSplitType read FSplitType write SetSplitType;
    property Transparent: Boolean read FTransparent write SetTransparent;
    property Visible;
    property OnShowPartition: TPartitionVisibilityEvent read FOnShowPartition write FOnShowPartition;
    property OnHidePartition: TPartitionVisibilityEvent read FOnHidePartition write FOnHidePartition;
  end;

procedure Register;

implementation

const
  SplitterD = 6;

constructor TSplitBox.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  SetBounds(Left, Top, 208, 120);
  FDivision := High(FDivision) div 2;
  ControlStyle := [csCaptureMouse, csClickEvents, csOpaque];
  Cursor := crHSplit;
  FMinSize1 := 32;
  FMinSize2 := 32;
  FPartition[1] := TSplitBoxPartition.Create(Self);
  FPartition[2] := TSplitBoxPartition.Create(Self);
  FPartition[1].Parent := Self;
  FPartition[2].Parent := Self;
  SetBevelInner(bvLowered);
  SetBevelOuter(bvNone);
end;

procedure TSplitBox.CreateParams(var Params: TCreateParams);
begin
  inherited CreateParams(Params);
  Params.Style := Params.Style or ws_Child or ws_ClipChildren;
end;

procedure TSplitBox.AlignControls(AControl: TControl; var Rect: TRect);
begin
  UpdatePartitions;
end;

procedure TSplitBox.AddPartition1(Control: TControl);
begin
  Control.Parent := FPartition[1];
end;

procedure TSplitBox.AddPartition2(Control: TControl);
begin
  Control.Parent := FPartition[2];
end;

procedure TSplitBox.CMDesignHitTest(var Msg: TCMDesignHitTest);
begin
  Msg.Result := 1;
end;
{
procedure TSplitBox.DefineProperties(Filer: TFiler);
begin
  inherited DefineProperties(Filer);
  Filer.DefineProperty('Division', ReadDivision, WriteDivision, True);
end;

procedure TSplitBox.ReadDivision(Reader: TReader);
begin
  FDivision := Reader.ReadInteger;
end;

procedure TSplitBox.WriteDivision(Writer: TWriter);
begin
  Writer.WriteInteger(FDivision);
end;
}
function TSplitBox.GetBevelInner: TPanelBevel;
begin
  Result := FPartition[1].BevelInner;
end;

function TSplitBox.GetBevelOuter: TPanelBevel;
begin
  Result := FPartition[1].BevelOuter;
end;

function TSplitBox.GetBevelWidth: TBevelWidth;
begin
  Result := FPartition[1].BevelWidth;
end;

function TSplitBox.GetBorderWidth: TBorderWidth;
begin
  Result := FPartition[1].BorderWidth;
end;

procedure TSplitBox.GetChildren(Proc: TGetChildProc; Root: TComponent);
begin
  FPartition[1].GetChildren(Proc, Root);
  FPartition[2].GetChildren(Proc, Root);
end;

function TSplitBox.GetDivision: Single;
begin
  Result := FDivision / High(FDivision);
end;

function TSplitBox.GetMinSizeLeftTop: NaturalNumber;
begin
  Result := FMinSize1;
end;

function TSplitBox.GetMinSizeRightBottom: NaturalNumber;
begin
  Result := FMinSize2;
end;

procedure TSplitBox.SetBevelInner(Value: TPanelBevel);
begin
  FPartition[1].BevelInner := Value;
  FPartition[2].BevelInner := Value;
  Invalidate;
end;

procedure TSplitBox.SetBevelOuter(Value: TPanelBevel);
begin
  FPartition[1].BevelOuter := Value;
  FPartition[2].BevelOuter := Value;
  Invalidate;
end;

procedure TSplitBox.SetBevelWidth(Value: TBevelWidth);
begin
  FPartition[1].BevelWidth := Value;
  FPartition[2].BevelWidth := Value;
  Invalidate;
end;

procedure TSplitBox.SetBorderWidth(Value: TBorderWidth);
begin
  FPartition[1].BorderWidth := Value;
  FPartition[2].BorderWidth := Value;
  Invalidate;
end;

procedure TSplitBox.SetDivision(Value: Single);
var
  D: Integer;
begin
  if Value < 0 then Value := 0 else
  if Value > 1 then Value := 1;
  D := Round(Value * High(FDivision));
  if D <> FDivision then
  begin
    FDivision := D;
    UpdatePartitions;
  end;
end;

procedure TSplitBox.SetMinSizeLeftTop(Value: NaturalNumber);
begin
  FMinSize1 := Value;
  UpdatePartitions;
end;

procedure TSplitBox.SetMinSizeRightBottom(Value: NaturalNumber);
begin
  FMinSize2 := Value;
  UpdatePartitions;
end;

procedure TSplitBox.Loaded;
var
  i, j, k: Integer;

  function NextControl: TControl;
  begin
    while (Controls[k] = FPartition[1])
       or (Controls[k] = FPartition[2]) do Inc(k);
    Result := Controls[k];
  end;

begin
  inherited Loaded;
  k := 0;
  j := ControlCount-3;
  for i := 0 to FPartition1Count-1 do
    NextControl.Parent := FPartition[1];
  for i := FPartition1Count to j do
    NextControl.Parent := FPartition[2];
end;

procedure TSplitBox.MouseDown;
begin
  inherited MouseDown(Button, Shift, X, Y);
  if Button = mbLeft then
  begin
    FSplitterDragging := True;
    FMovePos := Point(X, Y);
    SetWindowLong(Handle, gwl_Style, GetWindowLong(Handle, gwl_Style) and not ws_ClipChildren);
    InvertRect(Canvas.Handle, FSplitterRect);
  end;
end;

procedure TSplitBox.MouseMove;
var
  dX, dY: Integer;
begin
  if FSplitterDragging then
  begin
    if FSplitType = stHorizontal then
    begin
      dX := iMin(iMax(X-SplitterD shr 1, 0),
	  Width-SplitterD)-FSplitterRect.Left;
      if dX <> 0 then
      begin
	InvertRect(Canvas.Handle, FSplitterRect);
	OffsetRect(FSplitterRect, dX, 0);
	InvertRect(Canvas.Handle, FSplitterRect);
      end;
    end else
    begin
      dY := iMin(iMax(Y-SplitterD shr 1, 0),
	  Height-SplitterD)-FSplitterRect.Top;
      if dY <> 0 then
      begin
	InvertRect(Canvas.Handle, FSplitterRect);
	OffsetRect(FSplitterRect, 0, dY);
	InvertRect(Canvas.Handle, FSplitterRect);
      end;
    end;
    FMovePos := Point(X, Y);
  end;
  inherited MouseMove(Shift, X, Y);
end;

procedure TSplitBox.MouseUp;
begin
  if Button = mbLeft then
  if FSplitterDragging then
  begin
    FSplitterDragging := False;
    InvertRect(Canvas.Handle, FSplitterRect);
    SetWindowLong(Handle, gwl_Style, GetWindowLong(Handle, gwl_Style) or ws_ClipChildren);
    InvalidateRect(Handle, @FSplitterRect, True);
    UpdateDivision;
  end;
  inherited MouseUp(Button, Shift, X, Y);
end;

procedure TSplitBox.Paint;
var
  R: TRect;
  x, y: Integer;
begin
  R := FSplitterRect;
{ if (BevelOuter <> bvLowered)
  and (BevelInner <> bvLowered) then
    Frame3D(Canvas, R, clBtnHighLight, clBtnShadow, 1);}
  Canvas.Brush := Brush;
  Canvas.FillRect(R);
  with Canvas do
  if SplitType = stHorizontal then
  begin
    x := R.Left+1;
    y := R.Top+2;
    while y < R.Bottom-2 do
    begin
      Pixels[x, y] := clWhite;
      Inc(y, 3);
    end;
    Inc(x);
    y := R.Top+3;
    while y < R.Bottom-1 do
    begin
      Pixels[x, y] := cl3DDkShadow;
      Inc(y, 3);
    end;
    Inc(x);
    y := R.Top+1;
    while y < R.Bottom-3 do
    begin
      Pixels[x, y] := clWhite;
      Inc(y, 3);
    end;
    Inc(x);
    y := R.Top+2;
    while y < R.Bottom-2 do
    begin
      Pixels[x, y] := cl3DDkShadow;
      Inc(y, 3);
    end;
  end else
  begin
    x := R.Left+2;
    y := R.Top+1;
    while x < R.Right-2 do
    begin
      Pixels[x, y] := clWhite;
      Inc(x, 3);
    end;
    Inc(y);
    x := R.Left+3;
    while x < R.Right-1 do
    begin
      Pixels[x, y] := cl3DDkShadow;
      Inc(x, 3);
    end;
    Inc(y);
    x := R.Left+1;
    while x < R.Right-3 do
    begin
      Pixels[x, y] := clWhite;
      Inc(x, 3);
    end;
    Inc(y);
    x := R.Left+2;
    while x < R.Right-2 do
    begin
      Pixels[x, y] := cl3DDkShadow;
      Inc(x, 3);
    end;
  end;
end;

procedure TSplitBox.ReadState(Reader: TReader);
begin
  inherited ReadState(Reader);
  FPartition1Count := Reader.ReadInteger;
  UpdateTransparent;
end;

procedure TSplitBox.WriteState(Writer: TWriter);
begin
  inherited WriteState(Writer);
  Writer.WriteInteger(FPartition[1].ControlCount);
end;

procedure TSplitBox.SetPasteTarget(Value: TPasteTarget);
begin
  if FPasteTarget <> Value then
  begin
    FPasteTarget := Value;
    FPartition[Ord(Value)+1].SendToBack;
  end;
end;

procedure TSplitBox.SetSplitType(Value: TSplitType);
const
  SplitCursor: array[stHorizontal..stVertical]
	of TCursor = (crHSplit, crVSplit);
begin
  if FSplitType = Value then exit;
  FSplitType := Value;
  Cursor := SplitCursor[Value];
  Invalidate;
  UpdatePartitions;
end;

procedure TSplitBox.SetTransparent(Value: Boolean);
begin
  if Value = FTransparent then Exit;
  FTransparent := Value;
  if not (csReadingState in ControlState) then UpdateTransparent;
end;

procedure TSplitBox.UpdatePartitions;
begin
  if FSplitType = stHorizontal
    then UpdateHorz
    else UpdateVert;
end;

procedure TSplitBox.UpdateDivision;
begin
  if FSplitType = stHorizontal
    then FDivision := iMin(iMulDiv(FSplitterRect.Left+SplitterD shr 1, High(FDivision), Width), High(FDivision))
    else FDivision := iMin(iMulDiv(FSplitterRect.Top+SplitterD shr 1, High(FDivision), Height), High(FDivision));
{
  if FSplitType = stHorizontal
    then Division := (FSplitterRect.Left+SplitterD shr 1)/Width
    else Division := (FSplitterRect.Top+SplitterD shr 1)/Height;
}
  UpdatePartitions;
end;

procedure TSplitBox.ShowPartition(Index: TPartitionIndex);
begin
  if not FPartition[Index].Visible then
  begin
    FPartition[Index].Visible := True;
    if Assigned(FOnShowPartition) then FOnShowPartition(Self, Index);
  end;
end;

procedure TSplitBox.HidePartition(Index: TPartitionIndex);
begin
  if FPartition[Index].Visible then
  begin
    FPartition[Index].Visible := False;
    if Assigned(FOnHidePartition) then FOnHidePartition(Self, Index);
  end;
end;

procedure TSplitBox.UpdateHorz;
var
  D, X, Xr, Xl: Integer;
  W, H: Integer;
  R: TRect;
  procedure SizePartition1;
  begin
    FPartition[1].SetBounds(
	R.Left,
	R.Top,
	Xl,
	R.Bottom - R.Top);
    ShowPartition(1);
  end;
  procedure SizePartition2;
  begin
    FPartition[2].SetBounds(
	R.Left + Xr,
	R.Top,
	R.Right - R.Left- Xr,
	R.Bottom - R.Top);
    ShowPartition(2);
  end;
begin
  D := SplitterD;
  R := ClientRect;
  W := R.Right-R.Left;
  X := iMulDiv(W, FDivision, High(FDivision));
  if (X > W - FMinSize2) and not (csDesigning in ComponentState) then
  begin
    Xl := W-D;
    SizePartition1;
    HidePartition(2);
  end else
  if (X < FMinSize1) and not (csDesigning in ComponentState) then
  begin
    Xl := 0;
    Xr := D;
    HidePartition(1);
    SizePartition2;
  end else
  begin
    Xl := X-D shr 1;
    Xr := Xl + D;
    SizePartition1;
    SizePartition2;
  end;
  H := Height;
  with FSplitterRect do
  if (Xl <> Left)
  or (H  <> Bottom-Top) then
    FSplitterRect := Bounds(Xl, 0, D, H);
end;

procedure TSplitBox.UpdateVert;
var
  D, Y, Ya, Yb: Integer;
  W, H: Integer;
  R: TRect;
  procedure SizePartition1;
  begin
    FPartition[1].SetBounds(
	R.Left,
	R.Top,
	R.Right - R.Left,
	Ya);
    ShowPartition(1);
  end;
  procedure SizePartition2;
  begin
    FPartition[2].SetBounds(
	R.Left,
	R.Top + Yb,
	R.Right - R.Left,
	R.Bottom - R.Top- Yb);
    ShowPartition(2);
  end;
begin
  D := SplitterD;
  R := ClientRect;
  H := R.Bottom-R.Top;
  Y := iMulDiv(H, FDivision, High(FDivision));
  if (Y > H - FMinSize2) and not (csDesigning in ComponentState) then
  begin
    Ya := H-D;
    SizePartition1;
    HidePartition(2);
  end else
  if (Y < FMinSize1) and not (csDesigning in ComponentState) then
  begin
    Ya := 0;
    Yb := D;
    HidePartition(1);
    SizePartition2;
  end else
  begin
    Ya := Y-D shr 1;
    Yb := Ya + D;
    SizePartition1;
    SizePartition2;
  end;
  W := Width;
  with FSplitterRect do
  if (Ya <> Top)
  or (W  <> Right-Left) then
    FSplitterRect := Bounds(0, Ya, W, D);
end;

procedure TSplitBox.UpdateTransparent;
begin
  if FTransparent then
  begin
    FPartition[1].Brush.Style := bsClear;
    FPartition[2].Brush.Style := bsClear;
  end else
  begin
    FPartition[1].Brush.Style := bsSolid;
    FPartition[2].Brush.Style := bsSolid;
    FPartition[1].Invalidate;
    FPartition[2].Invalidate;
  end;
end;

procedure TSplitBox.WMSetCursor(var Msg: TWMSetCursor);
var
  Cur: HCursor;
begin
  Cur := 0;
  with Msg do
  begin
    if CursorWnd = Handle then
    case FSplitType of
      stHorizontal: Cur := Screen.Cursors[crHSplit];
      stVertical:   Cur := Screen.Cursors[crVSplit];
    end;
  end;
  if Cur <> 0 then SetCursor(Cur)
  else inherited;
end;

procedure TSplitBox.WndProc(var Message: TMessage);
begin
  with Message do
  case Msg of
    wm_KeyFirst..wm_KeyLast,
    wm_User..cm_Base,
    cm_Base+$100..cn_Base-1,
    cn_Base+wm_User..High(Integer):
      begin
	FPartition[1].Broadcast(Msg);
	FPartition[2].Broadcast(Msg);
      end;
    else inherited WndProc(Message);
  end;
end;

const
{$IFDEF Bilch}
  PalettePage = 'Homebrew';
{$ELSE}
  PalettePage = '3rdParty';
{$ENDIF}
procedure Register;
begin
  RegisterComponents(PalettePage, [TSplitBox]);
end;

end.

