(***********************************************************************
  Author:	Robert Rossmair
		Wagelwitz, Rodaer Str. 6
		04688 Mutzschen
		Germany

	e-mail:	Robert.Rossmair@t-online.de
	http://home.t-online.de/home/Robert.Rossmair/

  Module:	ZoomDisp

  Version:	1.0.12 (21-Nov-1997)

  Description:  ...

  History:
    27-Nov-97:	Version 1.1.1
	changed type to TWorldCoord for:
	TMarker.Position,
	TCustomZoomDisplay.DeviceX/Y arguments,
	TCustomZoomDisplay.MoveTo/LineTo arguments,
    26-Nov-97: Version 1.1.0
	added TCustomZoomDisplay.Error,
	TCustomZoomDisplay.PlotSmallintArray,
	TCustomZoomDisplay.PlotIntArray
    21-Nov-97: Version 1.0.12
	TCustomZoomDisplay.IterateMarkers
    20-Nov-97: Version 1.0.11
	added THorzMarker, TVertMarker classes
    17-Nov-97: Version 1.0.10
	added TMarker.Locked property
	added TCustomZoomDisplay.SetMarkerColorOf(),
    15-Nov-97: Version 1.0.9
	added TCustomZoomDisplay.SetMarkerHintOf(),
	      TCustomZoomDisplay.UpdateMarkerPosOf()
	      TMarker.SetOnGetposition()
    14-Nov-97: Version 1.0.8
	added TCustomZoomDisplay.DeleteMarkersMarking()
    10-Nov-97: Version 1.0.7
	added TMarker.Marked property
	published all Events inherited from TControl
    09-Nov-97: Version 1.0.6
	added TCustomZoomDisplay.FDoDrawMarkerPositions
    08-Nov-97: Version 1.0.5
	moved resource initialization (cursors)
	into initialization section
    01-Nov-97: Version 1.0.4
	added FRestoreColors and related stuff to equalize monochrome
	bitmap palette problems.
    29-Oct-97: Version 1.0.3
	added property TCustomZoomDisplay.Bitmap to make the paint
	buffer visible to the public.
    27-Oct-97: Version 1.0.2
	added condition "if FPosition <> Value then"
	to TMarker.SetPosition
    10-Oct-97: Version 1.0.1
	added condition "if not (csDestroying in ComponentState)" to
	TCustomZoomDisplay.WMSize. Without this condition an
	EAccessViolation is raised on program exit if:
	1. Two MDI child windows with TZoomDisplay controls placed on
	   each of them are open,
	2. one of these windows is maximized.
    17-Sep-97: Version 1.0.0

  Copyright  1997  Robert Rossmair

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program (file COPYING); if not, write to the Free
  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 ***********************************************************************)
unit ZoomDisp;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics,
  Controls, Forms, Dialogs, Math32, Colors, ExtCtrls, Buttons;

const
  crBase	= 10260;
  crCrossHairs	= crBase;
  crHand        = crBase+1;
  crGrip        = crBase+2;
  crZoomIn	= crBase+3;
  crZoomOut	= crBase+4;
type
  TWorldCoord = Double;
  TWorldPoint = record
    X, Y: TWorldCoord;
  end;

  EWindowSize = class(Exception);
  EZoomDisplayError = class(Exception);
  TCustomZoomDisplay = class;

  TPageFraction = 1..High(Smallint);
  TDisplayScrollBar = class(TPersistent)
  private
    FDisplay: TCustomZoomDisplay;
    FPageFraction: TPageFraction;
    FKind: TScrollBarKind;
    FVisible: Boolean;
    FPersistent: Boolean;
    FTracking: Boolean;
    constructor Create(ADisplay: TCustomZoomDisplay; AKind: TScrollBarKind);
    function GetMin: Integer;
    function GetMax: Integer;
    function GetMask(Flags: UInt): UInt;
    function GetPageSize: Integer;
    function GetPosition: Integer;
    function GetScrollPos: Integer;
    procedure SetPersistent(Value: Boolean);
    procedure SetPosition(Value: Integer);
    procedure SetVisible(Value: Boolean);
    procedure Update;
  public
    procedure Assign(Source: TPersistent); override;
    property Kind: TScrollBarKind read FKind;
    property Min: Integer read GetMin;
    property Max: Integer read GetMax;
    property PageSize: Integer read GetPageSize;
  published
    property PageFraction: TPageFraction read FPageFraction write FPageFraction default 10;
    property Persistent: Boolean read FPersistent write SetPersistent;
    property Position: Integer read GetPosition write SetPosition default 0;
    property Tracking: Boolean read FTracking write FTracking default False;
    property Visible: Boolean read FVisible write SetVisible default True;
  end;

  TMarker = class;
  TMarkerProc = procedure(Marker: TMarker; Param: DWord);

  TClipCodes = (ccCenter, ccLeft, ccRight, ccTop, ccBottom);
  TClipCode = set of TClipCodes;
  TZoomMode = (zmZoomOff, zmZoomIn, zmZoomOut);
  TZoomStartMode = (zmAutomatic, zmManual);
  TPosEvent = procedure(Sender: TObject; X, Y: Single) of object;
  TIntFunc = function(P: Pointer): Integer;

  TCustomZoomDisplay = class(TCustomControl)
  private
    FWorld: TRect;
    FWindow: TRect;
    FHorzScrollBar: TDisplayScrollBar;
    FVertScrollBar: TDisplayScrollBar;
    FXIncrement: Cardinal;
    FYIncrement: Cardinal;
    FMinWindowHeight: Integer;
    FMinWindowWidth: Integer;
    FMaxWindowHeight: Integer;
    FMaxWindowWidth: Integer;
    FXPos: Integer;
    FYPos: Integer;
    FHorzFormatStr: string;
    FVertFormatStr: string;
    FHorzScaleFactor: Single;
    FVertScaleFactor: Single;
    ViewPort: record
      W, H: integer;
    end;
    FPen: TPen;
    FBitmap: TBitmap;
    FMetafileCanvas: TMetafileCanvas;
    FOnChange: TNotifyEvent;
    FOnPaint: TNotifyEvent;
    FOnWorldPosChange: TPosEvent;
    FOnZoomMin: TNotifyEvent;
    FOnZoomMax: TNotifyEvent;
    FCurrentPos: TPoint;
    FDragPos: TPoint;
    FCurrentCode: TClipCode;
    FAutoScrollHorz: Boolean;
    FAutoScrollVert: Boolean;
    FZoomMode: TZoomMode;
    FZoomStartMode: TZoomStartMode;
    FZooming: Boolean;
    FInvalid: Boolean;
    FRestoreColors: Boolean;
    FDoDrawMarkerPositions: Boolean;
    procedure DoAutoScroll;
    procedure BoundWindow(var AWindow: TRect);
    procedure Change;
    function DevicePositionCode(P: TPoint): TClipCode;
    procedure DragLine(X, Y: Integer; State: TDragState);
    procedure DrawMarkerPositions;
    procedure Error(Text: string);
    function GetBitmap: TBitmap;
    function GetBuffered: Boolean;
    function GetCanvas: TCanvas;
    function GetWindowWidth: Integer;
    function GetWindowHeight: Integer;
    function GetWorldWidth: Integer;
    function GetWorldHeight: Integer;
    function GetWorldXStr(X: Integer): string;
    function GetWorldYStr(Y: Integer): string;
    procedure Invalidate; override;
    procedure MoveWindow(X, Y: Integer);
    procedure PenChanged(Sender: TObject);
    function PositionCode(P: TPoint): TClipCode;
    procedure HScroll(var Message: TWMScroll);
    procedure VScroll(var Message: TWMScroll);
    procedure ScrollTo(X, Y: Integer);
    procedure SetBuffered(Value: Boolean);
    procedure SetColor(Value: TColor);
    procedure SetPen(Value: TPen);
    procedure SetWindowSizeLimits(
	MinWidth, MinHeight,
	MaxWidth, MaxHeight: Integer);
    procedure SetMinWindowWidth(Value: Integer);
    procedure SetMinWindowHeight(Value: Integer);
    procedure SetMaxWindowWidth(Value: Integer);
    procedure SetMaxWindowHeight(Value: Integer);
    procedure SetHorzScrollBar(Value: TDisplayScrollBar);
    procedure SetVertScrollBar(Value: TDisplayScrollBar);
    procedure SetWindow(Value: TRect);
    procedure SetWindowLeft(Value: Integer);
    procedure SetWindowWidth(Value: Integer);
    procedure SetWindowTop(Value: Integer);
    procedure SetWindowHeight(Value: Integer);
    procedure SetWorld(const Value: TRect);
    procedure SetWorldLeft(Value: Integer);
    procedure SetWorldWidth(Value: Integer);
    procedure SetWorldTop(Value: Integer);
    procedure SetWorldHeight(Value: Integer);
    procedure UpdateScrollBars;
    function GetWorldPoint(X, Y: Integer): TPoint;
    procedure ContinueZoom(X, Y: Integer);
    procedure EndZoom;
    procedure DrawZoomBox;
    procedure UpdateMarkers;
    procedure WindowChanged;
    procedure WorldPosChanged(MouseX, MouseY: Integer);
    procedure CMDrag(var Message: TCMDrag); message CM_DRAG;
    procedure WMSize(var Message: TWMSize); message wm_Size;
    procedure WMHScroll(var Message: TWMHScroll); message WM_HSCROLL;
    procedure WMVScroll(var Message: TWMVScroll); message WM_VSCROLL;
  protected
    destructor Destroy; override;
    procedure CreateWnd; override;
    function DeviceX(X: TWorldCoord): Integer;
    function DeviceY(Y: TWorldCoord): Integer;
    procedure DragDrop(Source: TObject; X, Y: Integer); override;
    procedure DragOver(Source: TObject; X, Y: Integer; State: TDragState;
      var Accept: Boolean); override;
    function GetPalette: HPalette; override;
    function GetHorzPosStr(X: Single): string; virtual;
    function GetVertPosStr(Y: Single): string; virtual;
    procedure GetZoomWindow(var AWindow: TRect);
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);	override;
    procedure Paint; override;
    procedure RealizeZoom; virtual;
    procedure ScrollBy(Dx, Dy: Integer);
    procedure UpdateIncrements; virtual;
    function UpdateWorldWindow: boolean; virtual;
    procedure WndProc(var Message: TMessage); override;
    property AutoScrollHorz: Boolean read FAutoScrollHorz write FAutoScrollHorz;
    property AutoScrollVert: Boolean read FAutoScrollVert write FAutoScrollVert;
    property Bitmap: TBitmap read GetBitmap;
    property Buffered: Boolean read GetBuffered write SetBuffered;
    property Color write SetColor nodefault;
    property Cursor default crCrossHairs;
    property Width default 128;
    property Height default 128;
    property HorzFormatStr: string read FHorzFormatStr write FHorzFormatStr;
    property VertFormatStr: string read FVertFormatStr write FVertFormatStr;
    property HorzScaleFactor: Single read FHorzScaleFactor write FHorzScaleFactor;
    property VertScaleFactor: Single read FVertScaleFactor write FVertScaleFactor;
    property Pen: TPen read FPen write SetPen;
    property MinWindowWidth: Integer read FMinWindowWidth write SetMinWindowWidth default 4;		// Min(Right-Left+1)
    property MinWindowHeight: Integer read FMinWindowHeight write SetMinWindowHeight default 4;		// Min(Top-Bottom+1)
    property MaxWindowWidth: Integer read FMaxWindowWidth write SetMaxWindowWidth default High(Integer);// Max(Right-Left+1)
    property MaxWindowHeight: Integer read FMaxWindowHeight write SetMaxWindowHeight default 65536;	// Max(Top-Bottom+1)
    property HorzScrollBar: TDisplayScrollBar read FHorzScrollBar write SetHorzScrollBar;
    property VertScrollBar: TDisplayScrollBar read FVertScrollBar write SetVertScrollBar;
    property WindowLeft: Integer read FWindow.Left write SetWindowLeft;
    property WindowTop: Integer read FWindow.Top write SetWindowTop;
    property WindowWidth: Integer read GetWindowWidth write SetWindowWidth;
    property WindowHeight: Integer read GetWindowHeight write SetWindowHeight;
    property WorldLeft: Integer read FWorld.Left write SetWorldLeft;
    property WorldTop: Integer read FWorld.Top write SetWorldTop;
    property WorldWidth: Integer read GetWorldWidth write SetWorldWidth;
    property WorldHeight: Integer read GetWorldHeight write SetWorldHeight;
    property ZoomStartMode: TZoomStartMode read FZoomStartMode write FZoomStartMode;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
    property OnPaint: TNotifyEvent read FOnPaint write FOnPaint;
    property OnWorldPosChange: TPosEvent read FOnWorldPosChange write FOnWorldPosChange;
    property OnZoomMin: TNotifyEvent read FOnZoomMin write FOnZoomMin;
    property OnZoomMax: TNotifyEvent read FOnZoomMax write FOnZoomMax;
  public
    constructor Create(AOwner: TComponent); override;
    procedure BeginZoom(X, Y: Integer);
    procedure DeleteMarkersMarking(Marked: TObject);
    procedure IterateMarkers(Proc: TMarkerProc; Param: DWord);
    procedure DeviceLineTo(X, Y: Integer);
    procedure DeviceMoveTo(X, Y: Integer);
    procedure LineTo(X, Y: TWorldCoord);
    procedure MoveTo(X, Y: TWorldCoord);
    // Under construction
    procedure Plot(
	P: Pointer;
	PType: Word;
	PInc: Integer;
	WorldX1, WorldX2: Integer;
	Func: TIntFunc);
    procedure PlotSmallintArray(
	const Samples: array of Smallint;
	Offset: Longint;
	Channels: Integer;
	AChannel: Integer);
    procedure PlotIntArray(
	const Samples: array of Integer;
	Offset: Longint;
	Channels: Integer;
	AChannel: Integer);
    procedure SaveToFile(FileName: string);
    procedure SetMarkerColorOf(Marked: TObject; Color: TColor);
    procedure SetMarkerHintOf(Marked: TObject; Text: string);
    procedure UpdateMarkerPosOf(Marked: TObject);
    function WorldX(X: integer): Single;
    function WorldY(Y: integer): Single;
    property Canvas: TCanvas read GetCanvas;
    property HorzPosStr[X: Single]: string read GetHorzPosStr;
    property VertPosStr[Y: Single]: string read GetVertPosStr;
    property Window: TRect read FWindow write SetWindow;
    property World: TRect read FWorld write SetWorld;
    property WorldPoint[X, Y: Integer]: TPoint read GetWorldPoint;
    property WorldXStr[X: Integer]: string read GetWorldXStr;
    property WorldYStr[Y: Integer]: string read GetWorldYStr;
  published
    property ZoomMode: TZoomMode read FZoomMode write FZoomMode;
    property OnMouseDown;
  end;

  TZoomDisplay = class(TCustomZoomDisplay)
  published
    property Align;
    property AutoScrollHorz;
    property AutoScrollVert;
    property Bitmap;
    property Buffered;
    property Color;
    property Cursor;
    property Enabled;
    property Width;
    property Height;
    property HorzFormatStr;
    property VertFormatStr;
    property HorzScaleFactor;
    property VertScaleFactor;
    property ParentColor;
    property Pen;
    property PopupMenu;
    property MinWindowWidth;
    property MinWindowHeight;
    property MaxWindowWidth;
    property MaxWindowHeight;
    property HorzScrollBar;
    property VertScrollBar;
    property WindowLeft;
    property WindowTop;
    property WindowWidth;
    property WindowHeight;
    property WorldLeft;
    property WorldTop;
    property WorldWidth;
    property WorldHeight;
    property ZoomStartMode;
    property OnChange;
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnStartDrag;
    property OnEndDrag;
    property OnMouseMove;
    property OnMouseUp;
    property OnPaint;
    property OnWorldPosChange;
    property OnZoomMin;
    property OnZoomMax;
  end;

  TMarkerPosEvent = procedure(Sender: TObject; var Pos: TWorldCoord) of object;
  TMarkerKind = (mkMark, mkRangeStart, mkRangeEnd);

  TMarker = class(TGraphicControl)
  private
    FMarked: TObject;
    FID: Integer;
    FKind: TMarkerKind;
    FLocked: Boolean;
    FShowPosition: Boolean;
    FPosition: TWorldCoord;
    FRepeatTimer: TTimer;
    FHighLightColor: TColor;
    FShadowColor: TColor;
    FOnDelete: TNotifyEvent;
    FOnGetPosition: TMarkerPosEvent;
    FOnPosChange: TMarkerPosEvent;
    procedure DoEndDrag(Target: TObject; X, Y: Integer); override;
    procedure DoStartDrag(var DragObject: TDragObject); override;
    function GetPosition: TWorldCoord;
    procedure SetOnGetPosition(Value: TMarkerPosEvent);
    procedure SetPosition(Value: TWorldCoord);
    procedure SetShowPosition(Value: Boolean);
    procedure TimerExpired(Sender: TObject);
    procedure CMColorChanged(var Message: TMessage); message CM_COLORCHANGED;
  protected
    destructor Destroy; override;
    procedure AssignTo(Dest: TPersistent); override;
    function HotSpot: TPoint;
    procedure MouseDown(Button: TMouseButton;
	Shift: TShiftState; X, Y: Integer); override;
    procedure ParentInvalidate; virtual; abstract;
    { raises exception if AParent is no TCustomZoomDisplay }
    procedure SetParent(AParent: TWinControl); override;
    { Updates both world and device position }
    procedure UpdatePos; virtual; abstract;
  public
    constructor Create(AOwner: TComponent); override;
    procedure Delete;
    procedure UpdatePosition;
    property Color default clBtnFace;
    property Cursor default crHand;
    property DragCursor default crGrip;
    property ID: Integer read FID write FID;
    property Kind: TMarkerKind read FKind write FKind;
    property Locked: Boolean read FLocked write FLocked;
    property Marked: TObject read FMarked write FMarked;
    property PopupMenu;
    property Position: TWorldCoord read GetPosition write SetPosition;
    property ShowHint default True;
    property ShowPosition: Boolean read FShowPosition write SetShowPosition default True;
    property OnDelete: TNotifyEvent read FOnDelete write FOnDelete;
    property OnGetPosition: TMarkerPosEvent read FOnGetPosition write SetOnGetPosition;
    property OnPosChange: TMarkerPosEvent read FOnPosChange write FOnPosChange;
  end;

  THMarkerAlign = (alTop, alBottom);
  THorzMarker = class(TMarker)
  private
    FAlign: THMarkerAlign;
    procedure SetAlign(Value: THMarkerAlign);
  protected
    procedure Paint; override;
    procedure ParentInvalidate; override;
    procedure UpdatePos; override;
  public
    constructor Create(AOwner: TComponent); override;
    property Align: THMarkerAlign read FAlign write SetAlign;
  end;

  TVMarkerAlign = (alLeft, alRight);
  TVertMarker = class(TMarker)
  private
    FAlign: TVMarkerAlign;
    procedure SetAlign(Value: TVMarkerAlign);
  protected
    procedure Paint; override;
    procedure ParentInvalidate; override;
    procedure UpdatePos; override;
  public
    constructor Create(AOwner: TComponent); override;
    property Align: TVMarkerAlign read FAlign write SetAlign;
  end;

procedure Register;

implementation

{$R *.RES}
//uses Z_prof;

const
  InitRepeatPause = 400;  { pause before repeat timer (ms) }
  RepeatPause     =  50;  { repeat timer interval (ms)}
  SPlotSmallintArray = 'PlotSmallintArray: Length of array "Samples" < WindowWidth';
  SPlotIntArray = 'PlotIntArray: Length of array "Samples" < WindowWidth';

procedure VerifyWindowSize(const Value: TRect);
var
  StackTop: record end;
  Stack: record
    EBP: DWORD;
    ReturnAddress: Pointer;
  end absolute StackTop;
begin
  if (Value.Right <= Value.Left)
  or (Value.Top <= Value.Bottom) then
    raise EWindowSize.Create('Invalid window size') at Stack.ReturnAddress;
end;

constructor TMarker.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  Color := clBtnFace;
  Get3DColors(Color, FHighLightColor, FShadowColor);
  Cursor := crHand;
  DragCursor := crGrip;
  ShowHint := True;
  FShowPosition := True;
  Parent := (AOwner as TCustomZoomDisplay);
end;

destructor TMarker.Destroy;
begin
  ParentInvalidate;
  FRepeatTimer.Free;
  inherited Destroy;
end;

procedure TMarker.AssignTo(Dest: TPersistent);
var
  M: TMarker absolute Dest;
begin
  if Dest is TMarker then
  begin
    M.Color	:= Color;
    M.Cursor	:= Cursor;
    M.DragCursor := DragCursor;
    M.Hint	:= Hint;
    M.ID	:= ID;
    M.Kind	:= Kind;
    M.Locked	:= Locked;
    M.Marked	:= Marked;
    M.PopupMenu	:= PopupMenu;
    M.Position	:= Position;
    M.ShowHint	:= ShowHint;
    M.OnDelete	:= OnDelete;
    M.OnGetPosition := OnGetPosition;
    M.OnPosChange := OnPosChange;
  end
  else inherited AssignTo(Dest);
end;

procedure TMarker.CMColorChanged(var Message: TMessage);
begin
  Get3DColors(Color, FHighLightColor, FShadowColor);
  inherited;
end;

procedure TMarker.Delete;
begin
  if Assigned(FOnDelete) then FOnDelete(Self);
end;

procedure TMarker.DoStartDrag(var DragObject: TDragObject);
begin
  inherited;
  if FRepeatTimer = nil then
  begin
    FRepeatTimer := TTimer.Create(Self);
    FRepeatTimer.OnTimer := TimerExpired;
  end;
  FRepeatTimer.Interval := InitRepeatPause;
  FRepeatTimer.Enabled  := True;
end;

procedure TMarker.DoEndDrag(Target: TObject; X, Y: Integer);
begin
  inherited;
  FRepeatTimer.Enabled  := False;
end;

function TMarker.HotSpot: TPoint;
begin
  Result.X := Width div 2;
  Result.Y := Height div 2;
end;

procedure TMarker.MouseDown(
  Button: TMouseButton;
  Shift: TShiftState;
  X, Y: Integer);
begin
  if Button = mbLeft then if not Locked then BeginDrag(True);
  inherited MouseDown(Button, Shift, X, Y);
end;

function TMarker.GetPosition;
begin
  if Assigned(FOnGetPosition) then FOnGetPosition(Self, FPosition);
  Result := FPosition;
end;

procedure TMarker.SetParent(AParent: TWinControl);
begin
  inherited SetParent(AParent as TCustomZoomDisplay);
end;

procedure TMarker.SetPosition(Value: TWorldCoord);
begin
  if FPosition <> Value then
  begin
    FPosition := Value;
    if Assigned(FOnPosChange) then FOnPosChange(Self, FPosition);
    UpdatePosition;
  end;
end;

procedure TMarker.SetOnGetPosition(Value: TMarkerPosEvent);
begin
  if not CompareMem(@TMethod(Value), @TMethod(FOnGetPosition), SizeOf(TMethod)) then
  begin
    FOnGetPosition := Value;
    Parent.Invalidate;
  end;
end;

procedure TMarker.SetShowPosition(Value: Boolean);
begin
  if Value <> FShowPosition then
  begin
    FShowPosition := Value;
    ParentInvalidate;
  end;
end;

procedure TMarker.TimerExpired(Sender: TObject);
begin
  FRepeatTimer.Interval := RepeatPause;
  try
    (Parent as TCustomZoomDisplay).DoAutoScroll;
  except
    FRepeatTimer.Enabled := False;
    raise;
  end;
end;

procedure TMarker.UpdatePosition;
begin
  ParentInvalidate;
  UpdatePos;
  ParentInvalidate;
end;

constructor THorzMarker.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  SetBounds(0, 0, 9, 15);
end;

procedure THorzMarker.Paint;
type
  TBorder = array[0..4] of TPoint;
const
  Border: array[THMarkerAlign] of TBorder = ((
	(X: 4; Y: 14),
	(X: 8; Y: 10),
	(X: 8; Y: 0),
	(X: 0; Y: 0),
	(X: 0; Y: 10)),
	(
	(X: 0; Y: 14),
	(X: 8; Y: 14),
	(X: 8; Y: 4),
	(X: 4; Y: 0),
	(X: 0; Y: 4)));

  RangeLimitSymbols: array[THMarkerAlign, Boolean, 0..2] of TPoint = (
	((
	(X: 3; Y: 8),
	(X: 3; Y: 4),
	(X: 6; Y: 4)),
	(
	(X: 5; Y: 8),
	(X: 5; Y: 4),
	(X: 2; Y: 4))),
	((
	(X: 3; Y: 6),
	(X: 3; Y: 10),
	(X: 6; Y: 10)),
	(
	(X: 5; Y: 6),
	(X: 5; Y: 10),
	(X: 2; Y: 10))));
begin
  with Canvas do
  begin
    Pen.Color := FHighLightColor;
    Brush.Color := Color;
    Polygon(Border[FAlign]);
    Pen.Color := FShadowColor;
    PolyLine(Slice(Border[FAlign], 3));
    if FKind = mkMark then Exit;
    Pen.Mode := pmNot;
    PolyLine(RangeLimitSymbols[FAlign, FKind = mkRangeEnd]);
    Pen.Mode := pmCopy;
  end;
end;

procedure THorzMarker.ParentInvalidate;
var
  R: TRect;
begin
  if Parent <> nil then
  begin
    R.Left   := Left + HotSpot.X;
    R.Right  := R.Left+1;
    R.Top    := Height;
    R.Bottom := Parent.Height;
    InvalidateRect(Parent.Handle, @R, True);
  end;
end;

procedure THorzMarker.SetAlign(Value: THMarkerAlign);
begin
  if Value <> FAlign then
  begin
    FAlign := Value;
    if Value = alTop
      then SetBounds(Left, 0, Width, Height)
      else SetBounds(Left, Parent.Height-Height, Width, Height);
  end;
end;

procedure THorzMarker.UpdatePos;
begin
  if (Position > (Parent as TCustomZoomDisplay).Window.Right)
  or (Position < (Parent as TCustomZoomDisplay).Window.Left) then
  begin
    Visible := False;
  end else
  begin
    SetBounds((Parent as TCustomZoomDisplay).DeviceX(Position)-HotSpot.X, Top, Width, Height);
    Visible := True;
  end;
end;

constructor TVertMarker.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  SetBounds(0, 0, 15, 9);
end;

procedure TVertMarker.Paint;
type
  TBorder = array[0..4] of TPoint;
const
  Border: array[TVMarkerAlign] of TBorder = ((
	(X: 14; Y: 4),
	(X: 10; Y: 8),
	(X:  0; Y: 8),
	(X:  0; Y: 0),
	(X: 10; Y: 0)),
	(
	(X: 14; Y: 0),
	(X: 14; Y: 8),
	(X:  4; Y: 8),
	(X:  0; Y: 4),
	(X:  4; Y: 0)));

  RangeLimitSymbols: array[TVMarkerAlign, Boolean, 0..2] of TPoint = (
	((
	(X: 8; Y: 5),
	(X: 4; Y: 5),
	(X: 4; Y: 2)),
	(
	(X: 8; Y: 3),
	(X: 4; Y: 3),
	(X: 4; Y: 6))),
	((
	(X:  6; Y: 5),
	(X: 10; Y: 5),
	(X: 10; Y: 2)),
	(
	(X:  6; Y: 3),
	(X: 10; Y: 3),
	(X: 10; Y: 6))));
begin
  with Canvas do
  begin
    Pen.Color := FHighLightColor;
    Brush.Color := Color;
    Polygon(Border[FAlign]);
    Pen.Color := FShadowColor;
    PolyLine(Slice(Border[FAlign], 3));
    if FKind = mkMark then Exit;
    Pen.Mode := pmNot;
    PolyLine(RangeLimitSymbols[FAlign, FKind = mkRangeEnd]);
    Pen.Mode := pmCopy;
  end;
end;

procedure TVertMarker.ParentInvalidate;
var
  R: TRect;
begin
  if Parent <> nil then
  begin
    R.Top    := Top + HotSpot.Y;
    R.Bottom := R.Top+1;
    R.Left   := Width;
    R.Right  := Parent.Width;
    InvalidateRect(Parent.Handle, @R, True);
  end;
end;

procedure TVertMarker.SetAlign(Value: TVMarkerAlign);
begin
  if Value <> FAlign then
  begin
    FAlign := Value;
    if Value = alLeft
      then SetBounds(0, Top, Width, Height)
      else SetBounds(Parent.Width-Width, Top, Width, Height);
  end;
end;

procedure TVertMarker.UpdatePos;
begin
  if (Position > (Parent as TCustomZoomDisplay).Window.Right)
  or (Position < (Parent as TCustomZoomDisplay).Window.Left) then
  begin
    Visible := False;
  end else
  begin
    SetBounds(Left, (Parent as TCustomZoomDisplay).DeviceY(Position)-HotSpot.Y, Width, Height);
    Visible := True;
  end;
end;

constructor TDisplayScrollBar.Create(ADisplay: TCustomZoomDisplay;
  AKind: TScrollBarKind);
begin
  inherited Create;
  FDisplay := ADisplay;
  FKind := AKind;
  FPageFraction := 10;
  FVisible := True;
end;

procedure TDisplayScrollBar.Assign(Source: TPersistent);
begin
  if Source is TDisplayScrollBar then
  begin
    Visible := TDisplayScrollBar(Source).Visible;
    Exit;
  end;
  inherited Assign(Source);
end;

function TDisplayScrollBar.GetMask(Flags: UInt): UInt;
const
  ScrollInfoMask: array[Boolean] of UInt = (0, SIF_DISABLENOSCROLL);
begin
  Result := ScrollInfoMask[FPersistent] or Flags;
end;

function TDisplayScrollBar.GetMin: Integer;
begin
  if FKind = sbHorizontal
    then Result := FDisplay.FWorld.Left
    else Result := -FDisplay.FWorld.Top;
end;

function TDisplayScrollBar.GetMax: Integer;
begin
  with FDisplay do
  if FKind = sbHorizontal
    then Result := FWorld.Right
    else Result := -FWorld.Bottom;
end;

function TDisplayScrollBar.GetPageSize: Integer;
begin
  if FVisible then
    if FKind = sbHorizontal
      then Result := FDisplay.WindowWidth
      else Result := FDisplay.WindowHeight
  else Result := Max-Min+1;
end;

function TDisplayScrollBar.GetPosition: Integer;
begin
  with FDisplay do
  if FKind = sbHorizontal
    then Result := FXPos
    else Result := FYPos;
end;

function TDisplayScrollBar.GetScrollPos: Integer;
begin
  with FDisplay do
  if FKind = sbHorizontal
    then Result := FXPos
    else Result := 1-FYPos-PageSize;
end;

procedure TDisplayScrollBar.SetPersistent(Value: Boolean);
var
  Code: Integer;
  ScrollInfo: TScrollInfo;
begin
  if Value <> FPersistent then
  begin
    ScrollInfo.cbSize := SizeOf(ScrollInfo);
    ScrollInfo.fMask := SIF_ALL;
    Code := SB_HORZ;
    if FKind = sbVertical then Code := SB_VERT;
    GetScrollInfo(FDisplay.Handle, Code, ScrollInfo);
    FPersistent := Value;
    ScrollInfo.fMask := GetMask(ScrollInfo.fMask);
    SetScrollInfo(FDisplay.Handle, Code, ScrollInfo, True);
  end;
end;

procedure TDisplayScrollBar.SetPosition(Value: Integer);
begin
  with FDisplay do
  if FKind = sbHorizontal
    then ScrollTo(Value, FYPos)
    else ScrollTo(FXPos, Value);
end;

procedure TDisplayScrollBar.SetVisible(Value: Boolean);
begin
  if FVisible <> Value then
  begin
    FVisible := Value;
    FDisplay.UpdateScrollBars;
  end;
end;

procedure TDisplayScrollBar.Update;
var
  Code: Word;
  ScrollInfo: TScrollInfo;
begin
  ScrollInfo.cbSize := SizeOf(ScrollInfo);
  ScrollInfo.fMask := GetMask(SIF_ALL);
  if Kind = sbHorizontal
    then Code := SB_HORZ
    else Code := SB_VERT;
  ScrollInfo.nMin := Min;
  ScrollInfo.nMax := Max;
  ScrollInfo.nPage := PageSize;
  ScrollInfo.nPos := GetScrollPos;
  ScrollInfo.nTrackPos := GetScrollPos;
  SetScrollInfo(FDisplay.Handle, Code, ScrollInfo, True);
end;

{ TCustomZoomDisplay }

constructor TCustomZoomDisplay.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  ControlStyle := ControlStyle + [csOpaque];
  FPen := TPen.Create;
  FPen.OnChange := PenChanged;
  SetBounds(0, 0, 128, 128);
  FHorzScrollBar := TDisplayScrollBar.Create(Self, sbHorizontal);
  FVertScrollBar := TDisplayScrollBar.Create(Self, sbVertical);
  FHorzScaleFactor := 1.0;
  FVertScaleFactor := 1.0;
  FHorzFormatStr := '%0.2f';
  FVertFormatStr := '%0.2f';
  FWorld.Right := High(Integer)-1;
  FWorld.Top := High(Smallint);
  FWorld.Bottom := Low(SmallInt);
  FWindow := FWorld;
  FYPos := FWindow.Bottom;
  SetWindowSizeLimits(4, 4, FWorld.Right+1, FWorld.Top-FWorld.Bottom+1);
  Cursor := crCrossHairs;
  FDoDrawMarkerPositions := True;
end;

destructor TCustomZoomDisplay.Destroy;
begin
  FPen.Free;
  FBitmap.Free;
  FHorzScrollBar.Free;
  FVertScrollBar.Free;
  inherited Destroy;
end;

procedure TCustomZoomDisplay.SetColor(Value: TColor);
begin
  if Value <> Color then
  begin
    inherited Color := Value;
    if FBitmap.Monochrome then
      SetBitmapColors(FBitmap, [Color, FPen.Color], 0);
  end;
end;

procedure TCustomZoomDisplay.PenChanged(Sender: TObject);
begin
  if FBitmap.Monochrome then
    SetBitmapColors(FBitmap, [Color, FPen.Color], 0);
end;

procedure TCustomZoomDisplay.DoAutoScroll;
var
  CursorPos: TPoint;
  ClientRectangle: TRect;
  Dx, Dy: Integer;
begin
  if FAutoScrollVert
  or FAutoScrollHorz then
  begin
    GetCursorPos(CursorPos);
    CursorPos := ScreenToClient(CursorPos);
    ClientRectangle := GetClientRect;
    Dx := 0; Dy := 0;
    if FAutoScrollVert then
      if CursorPos.Y < 0 then
	Dy := iMin(-FYIncrement, iMax(-WindowHeight, (CursorPos.Y div 10) * FYIncrement)) else
      if CursorPos.Y > ClientRectangle.Bottom then
	Dy := iMax(FYIncrement, iMin(WindowHeight, ((CursorPos.Y - ClientRectangle.Bottom) div 10) * FYIncrement));
    if FAutoScrollHorz then
      if CursorPos.X < 0 then
	Dx := iMin(-FXIncrement, iMax(-WindowWidth, (CursorPos.X div 10) * FXIncrement)) else
      if CursorPos.X > ClientRectangle.Right then
	Dx := iMax(FXIncrement, iMin(WindowWidth, ((CursorPos.X - ClientRectangle.Right) div 10) * FXIncrement));
    ScrollBy(Dx, Dy);
  end;
end;

procedure TCustomZoomDisplay.CreateWnd;
begin
  inherited CreateWnd;
  UpdateScrollBars;
end;

procedure TCustomZoomDisplay.Change;
begin
  if Assigned(FOnChange) then FOnChange(Self);
end;

procedure TCustomZoomDisplay.Error(Text: string);
var
  StackTop: record end;
  Stack: record
    BPorEBP: Integer;
    ReturnAddress: Pointer;
  end absolute StackTop;
  N: string;
begin
  N := ClassName;
  raise EZoomDisplayError.CreateFmt('%s %s: %s',
    [PChar(@N[2]), { strip first character of ClassName }
    Name, Text]) at Stack.ReturnAddress;
end;

procedure TCustomZoomDisplay.Invalidate;
begin
  FInvalid := True;
  inherited Invalidate;
end;

function TCustomZoomDisplay.GetBitmap: TBitmap;
begin
  Buffered := True;
  Result := FBitmap;
end;

function TCustomZoomDisplay.GetBuffered: Boolean;
begin
  Result := FBitmap <> nil;
end;

procedure TCustomZoomDisplay.SetBuffered(Value: Boolean);
begin
  if Value <> Buffered then
  begin
    if Value then
    begin
      FBitmap := TBitmap.Create;
      FBitmap.Width := ViewPort.W;
      FBitmap.Height := ViewPort.H;
      FRestoreColors := True;
    end else
    begin
      FBitmap.Free;
      FBitmap := nil;
    end;
  end;
end;

function TCustomZoomDisplay.GetPalette: HPalette;
begin
  if Buffered
    then Result := FBitmap.Palette
    else Result := 0;
end;

procedure TCustomZoomDisplay.SetPen(Value: TPen);
begin
  FPen.Assign(Value);
end;

procedure TCustomZoomDisplay.SetHorzScrollBar(Value: TDisplayScrollBar);
begin
  FHorzScrollBar.Assign(Value);
end;

procedure TCustomZoomDisplay.SetVertScrollBar(Value: TDisplayScrollBar);
begin
  FVertScrollBar.Assign(Value);
end;

procedure TCustomZoomDisplay.UpdateScrollBars;
begin
  if HandleAllocated then
  begin
    FVertScrollBar.Update;
    FHorzScrollBar.Update;
  end;
end;

procedure TCustomZoomDisplay.SetWindow(Value: TRect);
begin
  VerifyWindowSize(Value);
  BoundWindow(Value);
  if not EqualRect(FWindow, Value) then
  begin
    FWindow := Value;
    FYPos := FWindow.Bottom;
    ScrollTo(FWindow.Left, FWindow.Bottom);
    Invalidate;
    WindowChanged;
  end;
end;

procedure TCustomZoomDisplay.SetWorld(const Value: TRect);
var
  W, H: Integer;
begin
  VerifyWindowSize(Value);
  if not EqualRect(Value, FWorld) then
  begin
    FWorld := Value;
    if not (csReading in ComponentState) then
    begin
      W := Value.Right-Value.Left+1;
      H := Value.Top-Value.Bottom+1;
      if MaxWindowWidth > W then MaxWindowWidth := W;
      if MaxWindowHeight > H then MaxWindowHeight := H;
      BoundWindow(FWindow);
    end;
  end;
end;

function TCustomZoomDisplay.GetWindowWidth: Integer;
begin
  Result := FWindow.Right-FWindow.Left+1;
end;

function TCustomZoomDisplay.GetWindowHeight: Integer;
begin
  Result := FWindow.Top-FWindow.Bottom+1;
end;

function TCustomZoomDisplay.GetWorldWidth: Integer;
begin
  Result := FWorld.Right-FWorld.Left+1;
end;

function TCustomZoomDisplay.GetWorldHeight: Integer;
begin
  Result := FWorld.Top-FWorld.Bottom+1;
end;

procedure TCustomZoomDisplay.SetWindowSizeLimits(
  MinWidth, MinHeight,
  MaxWidth, MaxHeight: Integer);
var
  MaxW, MaxH: Integer;
begin
  if not (csReading in ComponentState) then
  begin
    MaxW := FWorld.Right-FWorld.Left+1;
    MaxH := FWorld.Top-FWorld.Bottom+1;
    if MaxWidth > MaxW then MaxWidth := MaxW;
    if MaxHeight > MaxH then MaxHeight := MaxH;
    if MinWidth > MaxWidth then MinWidth := MaxWidth;
    if MinHeight > MaxHeight then MinHeight := MaxHeight;
  end;
  FMinWindowWidth := MinWidth;
  FMinWindowHeight := MinHeight;
  FMaxWindowWidth := MaxWidth;
  FMaxWindowHeight := MaxHeight;
  BoundWindow(FWindow);
  Invalidate;
end;

procedure TCustomZoomDisplay.SetMinWindowWidth(Value: Integer);
begin
  SetWindowSizeLimits(Value, MinWindowHeight, MaxWindowWidth, MaxWindowHeight);
end;

procedure TCustomZoomDisplay.SetMinWindowHeight(Value: Integer);
begin
  SetWindowSizeLimits(MinWindowWidth, Value, MaxWindowWidth, MaxWindowHeight);
end;

procedure TCustomZoomDisplay.SetMaxWindowWidth(Value: Integer);
begin
  SetWindowSizeLimits(MinWindowWidth, MinWindowHeight, Value, MaxWindowHeight);
end;

procedure TCustomZoomDisplay.SetMaxWindowHeight(Value: Integer);
begin
  SetWindowSizeLimits(MinWindowWidth, MinWindowHeight, MaxWindowWidth, Value);
end;

procedure TCustomZoomDisplay.SetWindowLeft(Value: Integer);
var
  W: TRect;
begin
  W := FWindow;
  W.Right := W.Right+Value-W.Left;
  W.Left := Value;
  SetWindow(W);
end;

procedure TCustomZoomDisplay.SetWindowWidth(Value: Integer);
var
  W: TRect;
begin
  W := FWindow;
  W.Right := W.Left+Value-1;
  SetWindow(W);
end;

procedure TCustomZoomDisplay.SetWindowTop(Value: Integer);
var
  W: TRect;
begin
  W := FWindow;
  W.Bottom := W.Bottom+Value-W.Top;
  W.Top := Value;
  SetWindow(W);
end;

procedure TCustomZoomDisplay.SetWindowHeight(Value: Integer);
var
  W: TRect;
begin
  W := FWindow;
  W.Bottom := W.Top-Value+1;
  SetWindow(W);
end;

procedure TCustomZoomDisplay.SetWorldLeft(Value: Integer);
var
  W: TRect;
begin
  W := FWorld;
  W.Right := W.Right+Value-W.Left;
  W.Left := Value;
  SetWorld(W);
end;

procedure TCustomZoomDisplay.SetWorldWidth(Value: Integer);
var
  W: TRect;
begin
  W := FWorld;
  W.Right := W.Left+Value-1;
  SetWorld(W);
end;

procedure TCustomZoomDisplay.SetWorldTop(Value: Integer);
var
  W: TRect;
begin
  W := FWorld;
  W.Bottom := W.Bottom+Value-W.Top;
  W.Top := Value;
  SetWorld(W);
end;

procedure TCustomZoomDisplay.SetWorldHeight(Value: Integer);
var
  W: TRect;
begin
  W := FWorld;
  W.Bottom := W.Top-Value+1;
  SetWorld(W);
end;

procedure TCustomZoomDisplay.BoundWindow(var AWindow: TRect);
var
  Dx, Dy: Integer;
begin
  if not (csLoading in ComponentState) then
  with AWindow do
  begin
    if Right-Left < MinWindowWidth-1 then Right := Left + MinWindowWidth-1 else
    if Right-Left > MaxWindowWidth-1 then Right := Left + MaxWindowWidth-1;
    if Top-Bottom < MinWindowHeight-1 then Top := Bottom + MinWindowHeight-1 else
    if Top-Bottom > MaxWindowHeight-1 then Top := Bottom + MaxWindowHeight-1;
    Dx := 0;
    Dy := 0;
    if Right > FWorld.Right then Dx := FWorld.Right-Right else
    if Left < FWorld.Left then Dx := FWorld.Left-Left;
    if Top > FWorld.Top then Dy := FWorld.Top-Top else
    if Bottom < FWorld.Bottom then Dy := FWorld.Bottom-Bottom;
    OffsetRect(AWindow, Dx, Dy);
  end;
end;

procedure TCustomZoomDisplay.DeviceLineTo(X, Y: Integer);
var
  PrevCode, Code1, Code2: TClipCode;
  P, Q: TPoint;

  procedure Clip(Code: TClipCode; var P: TPoint; Q: TPoint); near;
  begin
    if ccTop in Code then
    begin
      P.X := P.X + IMulDiv(-P.Y, Q.X-P.X, Q.Y-P.Y);
      P.Y := 0;
    end else
    if ccBottom in Code then
    begin
      P.X := P.X + IMulDiv(ViewPort.H-P.Y, Q.X-P.X, Q.Y-P.Y);
      P.Y := ViewPort.H;
    end else
    if ccRight in Code then
    begin
      P.Y := P.Y + IMulDiv(ViewPort.W-P.X, Q.Y-P.Y, Q.X-P.X);
      P.X := ViewPort.W;
    end else
    if ccLeft in Code then
    begin
      P.Y := P.Y + IMulDiv(-P.X, Q.Y-P.Y, Q.X-P.X);
      P.X := 0;
    end;
  end;

begin
  P := FCurrentPos;
  Q.X := X;
  Q.Y := Y;
  FCurrentPos := Q;
  PrevCode := FCurrentCode;
  Code1 := PrevCode;
  Code2 := DevicePositionCode(Q);
  FCurrentCode := Code2;
  repeat
    if Code1 * Code2 <> [ccCenter] then exit
    else if Code1 + Code2 = [ccCenter] then
    begin
      if PrevCode <> [ccCenter] then Canvas.MoveTo(P.X, P.Y);
      Canvas.LineTo(Q.X, Q.Y);
      exit;
    end;
    if Code1 = [ccCenter] then Clip(Code2, Q, P) else Clip(Code1, P, Q);
    Code1 := DevicePositionCode(P);
    Code2 := DevicePositionCode(Q);
  until False;
end;

procedure TCustomZoomDisplay.DeviceMoveTo(X, Y: Integer);
begin
  FCurrentPos.X := X;
  FCurrentPos.Y := Y;
  FCurrentCode := DevicePositionCode(FCurrentPos);
  if FCurrentCode = [ccCenter] then
    Canvas.MoveTo(FCurrentPos.X, FCurrentPos.Y);
end;

function TCustomZoomDisplay.DevicePositionCode(P: TPoint): TClipCode;
begin
  Result := [ccCenter];
  if P.Y < 0 then Include(Result, ccTop) else
  if P.Y > ViewPort.H then Include(Result, ccBottom);
  if P.X > ViewPort.W then Include(Result, ccRight) else
  if P.X < 0 then Include(Result, ccLeft);
end;

function TCustomZoomDisplay.DeviceX(X: TWorldCoord): Integer;
begin
//  Profile.Mark(1, True);
//  Result := IMulDiv(X-FWindow.Left, ViewPort.W, FWindow.Right-FWindow.Left);
  Result := Round((X-FWindow.Left)*ViewPort.W/(FWindow.Right-FWindow.Left));
//  Profile.Mark(1, False);
end;

function TCustomZoomDisplay.DeviceY(Y: TWorldCoord): Integer;
begin
//  Profile.Mark(2, True);
//  Result := IMulDiv(Y-FWindow.Top, ViewPort.H, FWindow.Bottom-FWindow.Top);
  Result := Round((Y-FWindow.Top)*ViewPort.H/(FWindow.Bottom-FWindow.Top));
//  Profile.Mark(2, False);
end;

function TCustomZoomDisplay.GetWorldPoint(X, Y: Integer): TPoint;
begin
  Result.X := FWindow.Left + IMulDiv(X, FWindow.Right-FWindow.Left, ViewPort.W);
  Result.Y := FWindow.Top  + IMulDiv(Y, FWindow.Bottom-FWindow.Top, ViewPort.H);
end;

procedure TCustomZoomDisplay.LineTo(X, Y: TWorldCoord);
begin
//  Profile.Mark(3, True);
  DeviceLineTo(DeviceX(X), DeviceY(Y));
//  Profile.Mark(3, False);
end;

procedure TCustomZoomDisplay.MoveTo(X, Y: TWorldCoord);
begin
  FCurrentPos.X := DeviceX(X);
  FCurrentPos.Y := DeviceY(Y);
  FCurrentCode := DevicePositionCode(FCurrentPos);
  if FCurrentCode = [ccCenter] then
    Canvas.MoveTo(FCurrentPos.X, FCurrentPos.Y);
end;

procedure TCustomZoomDisplay.MoveWindow(X, Y: Integer);
begin
  OffsetRect(FWindow, X-FWindow.Left, Y-FWindow.Bottom);
  BoundWindow(FWindow);
end;

function TCustomZoomDisplay.PositionCode(P: TPoint): TClipCode;
begin
  Result := [ccCenter];
  if P.Y > FWindow.Top    then Include(Result, ccTop) else
  if P.Y < FWindow.Bottom then Include(Result, ccBottom);
  if P.X > FWindow.Right  then Include(Result, ccRight) else
  if P.X < FWindow.Left   then Include(Result, ccLeft);
end;

procedure TCustomZoomDisplay.CMDrag(var Message: TCMDrag);
begin
  if Message.DragMessage = dmFindTarget
    then Message.Result := Longint(Self)
    else inherited;
end;

procedure TCustomZoomDisplay.DragDrop(
  Source: TObject;
  X, Y: Integer);
begin
  if Source is THorzMarker
    then TMarker(Source).Position := WorldX(X) else
  if Source is TVertMarker
    then TMarker(Source).Position := WorldY(Y)
  else inherited;
end;

procedure TCustomZoomDisplay.DragLine(
  X, Y: Integer;
  State: TDragState);
var
  C: TCanvas;
begin
  C := inherited Canvas;
  C.Pen.Mode := pmNot;
  if not (State = dsDragEnter) then
  begin
    C.MoveTo(FDragPos.X, 0);
    C.LineTo(FDragPos.X, Height);
  end;
  FDragPos.X := X;
  FDragPos.Y := Y;
  if not (State = dsDragLeave) then
  begin
    C.MoveTo(X, 0);
    C.LineTo(X, Height);
  end;
end;

procedure TCustomZoomDisplay.DragOver(
  Source: TObject;
  X, Y: Integer;
  State: TDragState;
  var Accept: Boolean);
begin
  inherited;
  WorldPosChanged(X, Y);
  if Accept and (Source is TMarker) then DragLine(X, Y, State);
end;

procedure TCustomZoomDisplay.MouseMove(Shift: TShiftState; X, Y: Integer);
begin
  inherited;
  WorldPosChanged(X, Y);
  if FZooming then ContinueZoom(X, Y);
end;

procedure TCustomZoomDisplay.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  if Button = mbLeft then
  if not (csDesigning in ComponentState) then
  if FZoomMode <> zmZoomOff then
  if FZoomStartMode = zmAutomatic then
  begin
    BeginZoom(X, Y);
    Exit;
  end;
  inherited MouseDown(Button, Shift, X, Y);
end;

procedure TCustomZoomDisplay.MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  if Button = mbLeft then if FZooming then EndZoom;
  inherited;
end;

procedure TCustomZoomDisplay.WndProc(var Message: TMessage);
begin
  case Message.Msg of
    wm_EraseBkgnd:
	begin
	  Message.Result := 1;
	  Exit;
	end;
    wm_Paint: UpdateWorldWindow;
  end;
  inherited WndProc(Message);
end;

procedure TCustomZoomDisplay.DrawMarkerPositions;
var
  i, X, Y: Integer;
  M: TMarker;
begin
  if not FDoDrawMarkerPositions then Exit;
  Canvas.Pen.Mode := pmNot;
  for i := 0 to ControlCount-1 do
  begin
    M := TMarker(Controls[i]);
    if M.Visible and M.ShowPosition then
    if M is THorzMarker then
    begin
      X := DeviceX(M.Position);
      Canvas.MoveTo(X, 0);
      Canvas.LineTo(X, ViewPort.H+1);
    end else
    if M is TVertMarker then
    begin
      Y := DeviceY(M.Position);
      Canvas.MoveTo(0, Y);
      Canvas.LineTo(ViewPort.W+1, Y);
    end;
  end;
end;

procedure TCustomZoomDisplay.Paint;
begin
  if csDesigning in ComponentState then
  begin
    inherited Canvas.Brush := Brush;
    with inherited Canvas do
    begin
      Pen.Style := psDash;
      Rectangle(0, 0, Width, Height);
    end;
  end else
  begin
    Canvas.Pen := FPen;
    Canvas.Brush := Brush;
    Canvas.FillRect(ClientRect);
    if FRestoreColors then
    begin
      if FBitmap.Monochrome then
	SetBitmapColors(FBitmap, [Color, FPen.Color], 0);
      FRestoreColors := False;
    end;
    if Assigned(FOnPaint) then FOnPaint(Self);
    DrawMarkerPositions;
    if FBitmap <> nil then inherited Canvas.Draw(0, 0, FBitmap);
  end;
end;

function TCustomZoomDisplay.GetCanvas: TCanvas;
begin
  Result := FMetafileCanvas;
  if Result = nil then
  if Buffered
    then Result := FBitmap.Canvas
    else Result := inherited Canvas;
end;

var
  ZoomBox: TSmallRect;

procedure TCustomZoomDisplay.DrawZoomBox;
var
  R, B: Integer;
begin
// Rectangle doesn't include the right column and bottom row of pixels.
  with ZoomBox do
  begin
    R := Right;
    if R >= Left then Inc(R);
    B := Bottom;
    if B >= Top then Inc(B);
    inherited Canvas.Rectangle(Left, Top, R, B);
  end;
end;

procedure TCustomZoomDisplay.BeginZoom(X, Y: Integer);
var
  R: TRect;
begin
  if FZooming then Exit;
  FZooming := True;
  Update;
  ZoomBox.Left	  := X;
  ZoomBox.Top	  := Y;
  ZoomBox.Right  := X;
  ZoomBox.Bottom := Y;
  R := GetClientRect;
  MapWindowPoints(Handle, hWnd_Desktop, R, 2);
  ClipCursor(@R);
  MouseCapture := True;
  with inherited Canvas do
  begin
    Brush.Style := bsClear;
    Pen.Style := psDot;
    Pen.Color := clWhite;
    Pen.Mode := pmXor;
  end;
  DrawZoomBox;
end;

procedure TCustomZoomDisplay.ContinueZoom(X, Y: Integer);
begin
  DrawZoomBox;
  ZoomBox.Right  := X;
  ZoomBox.Bottom := Y;
  DrawZoomBox;
end;

procedure TCustomZoomDisplay.EndZoom;
begin
  DrawZoomBox;
  Dec(ZoomBox.Right);
  Dec(ZoomBox.Bottom);
  ClipCursor(nil);
  FZooming := False;
  MouseCapture := False;
  RealizeZoom;
end;

procedure TCustomZoomDisplay.GetZoomWindow(var AWindow: TRect);
var
  dX, dY, Temp: Integer;
  dZoomBoxX: integer;
  dZoomBoxY: integer;
  P, Q: TPoint;
begin
  with ZoomBox do
  begin
    if Right < Left then
    begin
      Temp := Left;
      Left := Right;
      Right := Temp;
    end else
    if Right = Left then Right := Left + 1;
    if Bottom < Top then
    begin
      Temp := Top;
      Top := Bottom;
      Bottom := Temp;
    end else
    if Bottom = Top then Bottom := Top + 1;
  end;
  with AWindow do
  if FZoomMode = zmZoomOut then
  begin
    dX := FWindow.Right - FWindow.Left;
    dY := FWindow.Bottom - FWindow.Top;
    dZoomBoxX := ZoomBox.Right-ZoomBox.Left;
    dZoomBoxY := ZoomBox.Bottom-ZoomBox.Top;
    try
      Left   := FWindow.Left - IMulDiv(ZoomBox.Left, dX, dZoomBoxX);
      Right  := FWindow.Left + IMulDiv(ViewPort.W,    dX, dZoomBoxX);
    except
      on EIntError do
	begin
	  AWindow.Left := FWorld.Left;
	  AWindow.Right := FWorld.Right;
	end;
    end;
    try
      Top    := FWindow.Top  - IMulDiv(ZoomBox.Top,  dY, dZoomBoxY);
      Bottom := FWindow.Top  + IMulDiv(ViewPort.H,    dY, dZoomBoxY);
    except
      on EIntError do
	begin
	  AWindow.Top := FWorld.Top;
	  AWindow.Bottom := FWorld.Bottom;
	end;
    end;
  end else
  begin
    with ZoomBox do
    begin
      P := WorldPoint[Left, Top];
      Q := WorldPoint[Right, Bottom];
    end;
    Left   := P.X;
    Top    := P.Y;
    Right  := Q.X;
    Bottom := Q.Y;
  end;
  BoundWindow(AWindow);
end;

procedure TCustomZoomDisplay.RealizeZoom;
var
  ZoomWindow: TRect;
begin
  GetZoomWindow(ZoomWindow);
  if not EqualRect(FWindow, ZoomWindow) then
    SetWindow(ZoomWindow);
end;

procedure TCustomZoomDisplay.WMSize(var Message: TWMSize);
begin
  ViewPort.W := Message.Width-1;
  ViewPort.H := Message.Height-1;
  if FBitmap <> nil then
  begin
    if FBitmap.Width <> Message.Width then
    begin
      FBitmap.Width := Message.Width;
      FRestoreColors := True;
    end;
    if FBitmap.Height <> Message.Height then
    begin
      FBitmap.Height := Message.Height;
      FRestoreColors := True;
    end;
  end;
  inherited;
  if not (csDestroying in ComponentState) then UpdateMarkers;
end;

procedure TCustomZoomDisplay.WMHScroll(var Message: TWMHScroll);
begin
  if Message.ScrollBar = 0 then HScroll(Message) else inherited;
end;

procedure TCustomZoomDisplay.WMVScroll(var Message: TWMVScroll);
begin
  if Message.ScrollBar = 0 then VScroll(Message) else inherited;
end;

procedure TCustomZoomDisplay.ScrollBy(Dx, Dy: Integer);
begin
  ScrollTo(FWindow.Left + Dx, FWindow.Bottom + Dy);
end;

procedure TCustomZoomDisplay.ScrollTo(X, Y: Integer);
begin
  X := iMax(FWorld.Left, iMin(X, FWorld.Right-(FWindow.Right-FWindow.Left)));
  Y := iMax(FWorld.Bottom, iMin(Y, FWorld.Top-(FWindow.Top-FWindow.Bottom)));
  if (X <> FXPos) or (Y <> FYPos) then
  begin
    FXPos := X;
    FYPos := Y;
    Invalidate;
  end;
end;

procedure TCustomZoomDisplay.HScroll(var Message: TWMScroll);
var
  ScrollInfo: TScrollInfo;
begin
  ScrollInfo.cbSize := SizeOf(ScrollInfo);
  ScrollInfo.fMask := HorzScrollBar.GetMask(SIF_ALL);
  GetScrollInfo(Handle, SB_HORZ, ScrollInfo);
  with Message do
    case ScrollCode of
      SB_LINEUP: ScrollBy(-FXIncrement, 0);
      SB_LINEDOWN: ScrollBy(FXIncrement, 0);
      SB_PAGEUP: ScrollBy(-WindowWidth, 0);
      SB_PAGEDOWN: ScrollBy(WindowWidth, 0);
      SB_THUMBPOSITION: ScrollTo(ScrollInfo.nPos, FYPos);
      SB_THUMBTRACK:
	with FHorzScrollBar do
	begin
	  if Tracking then ScrollTo(ScrollInfo.nTrackPos, FYPos);
	  ScrollInfo.nPos := ScrollInfo.nTrackPos;
	  SetScrollInfo(Handle, SB_HORZ, ScrollInfo, False);
	end;
      SB_TOP: ScrollTo(FWorld.Left, FYPos);
      SB_BOTTOM: ScrollTo(FWorld.Right-WindowWidth, FYPos);
      SB_ENDSCROLL: ;
    end;
end;

procedure TCustomZoomDisplay.VScroll(var Message: TWMScroll);
var
  ScrollInfo: TScrollInfo;
begin
  ScrollInfo.cbSize := SizeOf(ScrollInfo);
  ScrollInfo.fMask := VertScrollBar.GetMask(SIF_ALL);
  GetScrollInfo(Handle, SB_VERT, ScrollInfo);
  with Message do
    case ScrollCode of
      SB_LINEUP: ScrollBy(0, FYIncrement);
      SB_LINEDOWN: ScrollBy(0, -FYIncrement);
      SB_PAGEUP: ScrollBy(0, WindowHeight);
      SB_PAGEDOWN: ScrollBy(0, -WindowHeight);
      SB_THUMBPOSITION: ScrollTo(FXPos, -WindowHeight-ScrollInfo.nPos);
      SB_THUMBTRACK:
	with FVertScrollBar do
	begin
	  if Tracking then ScrollTo(FXPos, -WindowHeight-ScrollInfo.nTrackPos);
	  ScrollInfo.nPos := ScrollInfo.nTrackPos;
	  SetScrollInfo(Handle, SB_VERT, ScrollInfo, False);
	end;
      SB_TOP: ScrollTo(FXPos, FWorld.Top-WindowHeight);
      SB_BOTTOM: ScrollTo(FXPos, FWorld.Bottom);
      SB_ENDSCROLL: ;
    end;
end;

procedure TCustomZoomDisplay.UpdateIncrements;
begin
  FXIncrement := iMax(WindowWidth div FHorzScrollBar.PageFraction, 1);
  FYIncrement := iMax(WindowHeight div FVertScrollBar.PageFraction, 1);
end;

procedure TCustomZoomDisplay.IterateMarkers(Proc: TMarkerProc; Param: DWord);
var
  i: Integer;
  M: TMarker;
begin
  for i := 0 to ControlCount-1 do
  begin
    M := TMarker(Controls[i]);
    if M is TMarker then Proc(M, Param);
  end;
end;

procedure TCustomZoomDisplay.UpdateMarkers;
var
  i: Integer;
  M: TMarker;
begin
  for i := 0 to ControlCount-1 do
  begin
    M := TMarker(Controls[i]);
    if M is TMarker then M.UpdatePos;
  end;
end;

procedure TCustomZoomDisplay.UpdateMarkerPosOf(Marked: TObject);
var
  i: Integer;
  M: TMarker;
  Flag: Boolean;
begin
  Flag := False;
  for i := 0 to ControlCount-1 do
  begin
    M := TMarker(Controls[i]);
    if M is TMarker then if M.Marked = Marked then
    begin
      M.UpdatePos;
      Flag := True;
    end;
  end;
  if Flag then Invalidate;
end;

procedure TCustomZoomDisplay.SetMarkerColorOf(Marked: TObject; Color: TColor);
var
  i: Integer;
  M: TMarker;
begin
  for i := 0 to ControlCount-1 do
  begin
    M := TMarker(Controls[i]);
    if M is TMarker then if M.Marked = Marked then M.Color := Color;
  end;
end;

procedure TCustomZoomDisplay.SetMarkerHintOf(Marked: TObject; Text: string);
var
  i: Integer;
  M: TMarker;
begin
  for i := 0 to ControlCount-1 do
  begin
    M := TMarker(Controls[i]);
    if M is TMarker then if M.Marked = Marked then M.Hint := Text;
  end;
end;

procedure TCustomZoomDisplay.DeleteMarkersMarking(Marked: TObject);
var
  i: Integer;
  M: TMarker;
begin
  i := 0;
  while i < ControlCount do
  begin
    M := TMarker(Controls[i]);
    if (M is TMarker) and (M.Marked = Marked)
      then M.Free
      else Inc(i);
  end;
end;

function TCustomZoomDisplay.UpdateWorldWindow: Boolean;
begin
  Result := FInvalid;
  if Result then
  begin
    MoveWindow(FXPos, FYPos);
    UpdateIncrements;
    UpdateScrollBars;
    UpdateMarkers;
    Change;
    FInvalid := False;
  end;
end;

function TCustomZoomDisplay.WorldX(X: Integer): Single;
begin
  WorldX := FWindow.Left + (FWindow.Right-FWindow.Left) * (X / ViewPort.W);
end;

function TCustomZoomDisplay.WorldY(Y: Integer): Single;
begin
  WorldY := FWindow.Top + (FWindow.Bottom-FWindow.Top) * (Y / ViewPort.H);
end;

function TCustomZoomDisplay.GetHorzPosStr(X: Single): string;
begin
  Result := Format(HorzFormatStr, [FHorzScaleFactor * X]);
end;

function TCustomZoomDisplay.GetVertPosStr(Y: Single): string;
begin
  Result := Format(VertFormatStr, [FVertScaleFactor * Y]);
end;

function TCustomZoomDisplay.GetWorldXStr(X: Integer): string;
begin
  Result := GetHorzPosStr(WorldX(X));
end;

function TCustomZoomDisplay.GetWorldYStr(Y: Integer): string;
begin
  Result := GetVertPosStr(WorldY(Y));
end;

procedure TCustomZoomDisplay.WorldPosChanged(MouseX, MouseY: Integer);
begin
  if Assigned(FOnWorldPosChange) then FOnWorldPosChange(Self, WorldX(MouseX), WorldY(MouseY));
end;

procedure TCustomZoomDisplay.WindowChanged;
begin
  if Assigned(FOnZoomMax) then
  if (WindowWidth = MinWindowWidth) and (WindowHeight = MinWindowHeight) then
    FOnZoomMax(Self);
  if Assigned(FOnZoomMin) then
  if (WindowWidth = MaxWindowWidth) and (WindowHeight = MaxWindowHeight) then
    FOnZoomMin(Self);
end;

procedure TCustomZoomDisplay.SaveToFile(FileName: string);
var
  M: TMetafile;
begin
  M := TMetafile.Create;
  try
    M.Width := ViewPort.W;
    M.Height := ViewPort.H;
    FMetaFileCanvas := TMetaFileCanvas.Create(M, inherited Canvas.Handle);
    FDoDrawMarkerPositions := False;
    try
      Paint;
    finally
      FDoDrawMarkerPositions := True;
      FMetafileCanvas.Free;
      FMetafileCanvas := nil;
    end;
    M.SaveToFile(FileName);
  finally
    M.Free;
  end;
end;

function BytesPerScanline(PixelsPerScanline, BitsPerPixel, Alignment: Longint): Longint;
begin
  Dec(Alignment);
  Result := ((PixelsPerScanline * BitsPerPixel) + Alignment) and not Alignment;
  Result := Result div 8;
end;

procedure TCustomZoomDisplay.PlotSmallintArray(
  const Samples: array of Smallint;
  Offset: Longint;
  Channels: Integer;
  AChannel: Integer);
{ Die Idee hinter der Routine fr den Fall, da das
  Weltfenster breiter ist als das Bildschirmfenster,
  ist folgende:
  Wenn mehr als eine Welt-X-Koordinate (Wx) auf
  eine Bildschirm-X-Koordinate (X) entfllt, gehe
  man so vor:
1. Aufeinanderfolgende Samples, die sich in der
   horizontalen Bildschirmkoordinate unterscheiden,
   werden durch eine Linie verbunden;
2. Haben mehrere Samples dieselbe horizontale
   Bildschirmkoordinate, so werden Minimum und
   Maximum ihrer Amplituden festgestellt und
   durch eine senkrechte Linie dargestellt.
3. Das legt eine Verfeinerung nahe: Die Hlfte
   der Linie von 1. jeweils der entsprechenden
   senkrechten aus 2. zuschlagen. }
type
  TSample = Smallint;
  PSample = ^TSample;
var
  AmpMin: Integer;
  AmpMax: Integer;
  X, Y, W, H: Integer;
  Ymin: Integer;
  Ymax: Integer;
  YminPrev: Integer;
  YmaxPrev: Integer;
  WH: Integer;
  WX: Integer;
  WY: Integer;
  WX1: Integer;
  WX2: Integer;
  WY2: Integer;
  WXmax: Integer;
  WXBufEnd: Integer;
  Sample: PSample;
  BufRem: Integer;
  Bits: PByteArray;
  ScanLineW: Integer;
  SampleCount: Integer;

  procedure VerticalLine(X, Ymin, Ymax: Integer);
  var
    Y: Integer;
    Mask: DWORD;
    ScanLineY: PByteArray;
  begin
    if X < 0 then Exit;
    Y := iMin(H-1, Ymin);
    if Y < 0 then Exit;
    Mask := $80 shr (X and 7);
    ScanLineY := PByteArray(Integer(Bits)+ScanLineW*(H-1-Y)); // PByteArray(Bitmap.ScanLine[Y]);
    for Y := Y downto iMax(0, Ymax) do
    begin
      ScanLineY^[X shr 3] := ScanLineY^[X shr 3] or Mask;
      Inc(PByte(ScanLineY), ScanLineW);
    end;
  end;

  procedure GetMinMax(Start, Stop: Integer);	assembler;
  {
  var
    i: Integer;
  begin
    for i := Start to Stop do
    begin
       WY := Sample^;
       if WY < AmpMin then AmpMin := WY;
       if WY > AmpMax then AmpMax := WY;
       Inc(Sample, Channels);
    end;
  end;
    // Min in bx, Max in dx
  }
  asm
	mov	ecx, Stop
	sub	ecx, Start
	jl	@end
	inc	ecx
	push	ebx
	push	esi
	push	edi
	mov	esi, Sample
	mov	edi, Channels
	sal	edi, 1	// edi := Channels * SizeOf(Sample^)
	mov	bx, WORD PTR AmpMin
	mov	dx, WORD PTR AmpMax
  @GetNext:
	mov	ax, WORD PTR [esi]
	add	esi, edi
	cmp	bx, ax
	jng	@GetMax
	mov	bx, ax
  @GetMax:
	cmp	dx, ax
	jng	@1
	dec	ecx
	jnz	@GetNext
	jmp	@exit
  @1:	mov	dx, ax
	dec     ecx
	jnz	@GetNext
  @exit:
	mov	Sample, esi
	cwde
	mov	WY, eax
	movsx	eax, bx
	mov	AmpMin, eax
	movsx	eax, dx
	mov	AmpMax, eax
	pop	edi
	pop	esi
	pop	ebx
  @end:
  end;

begin
  with Bitmap do
  begin
    if PixelFormat <> pf1bit then
    begin
      PixelFormat := pf1bit;
      SetBitmapColors(FBitmap, [Color, FPen.Color], 0);
    end;
    ScanLineW := BytesPerScanline(Width, 1, 32);
    Bits := ScanLine[Height-1];
    FillChar(Bits^, Height * ScanLineW, 0);
  end;
  WX1 := Window.Left;
  WX2 := Window.Right;
  Sample := PSample(@Samples);
  SampleCount := (High(Samples)-Low(Samples)+1) div Channels;
  if SampleCount < WX2-WX1+1 then Error(SPlotSmallintArray);
  BufRem := iMin(WX2-WX1+1, SampleCount-Offset);
  Inc(Sample, Offset * Channels + AChannel);
  WX := WX1;
  W := ClientWidth;
  if W >= WX2-WX1 then
  begin
    MoveTo(WX, Sample^);
    WXmax := WX + Pred(BufRem);
    for WX := Succ(WX) to WXmax do
    begin
      Inc(Sample, Channels);
      LineTo(WX, Sample^);
    end;
    Sample := PSample(@Samples);
    Inc(Sample, AChannel);
    BufRem := WX2-WX1+1-BufRem;
    for WX := Succ(WXmax) to WXmax + Pred(BufRem) do
    begin
      LineTo(WX, Sample^);
      Inc(Sample, Channels);
    end;
  end else
  begin
    H := ClientHeight;
    WY2 := Window.Top;
    WH := Window.Bottom-WY2;
    WY := Sample^;
    YMinPrev := IMulDiv(WY-WY2, H, WH);
    YmaxPrev := YminPrev;
    for X := 0 to W do
    begin
      Y := Trunc(((WY+Sample^)/2-WY2)*H/WH);
      AmpMin := High(TSample);
      AmpMax := Low(TSample);
      WXmax := iMin(WX2, WX1 + iMulDiv(X+1, WX2-WX1, W));
      WXBufEnd := WX+BufRem-1;
      if WXmax > WXBufEnd then
      begin
	GetMinMax(WX, WXBufEnd);
	WX := WXBufEnd+1;
	BufRem := iMin(WX2-WX+1, SampleCount);
	Sample := PSample(@Samples);
	Inc(Sample, AChannel);
      end;
      Dec(BufRem, WXmax-WX+1);
      GetMinMax(WX, WXmax);
      WX := WXmax+1;
      Ymin := IMulDiv(AmpMin-WY2, H, WH);
      Ymax := IMulDiv(AmpMax-WY2, H, WH);
      if Ymin < YmaxPrev-1 then
      begin
	Ymin := iMax(Ymin, Y);
	YmaxPrev := iMin(YmaxPrev, (Ymin+1));
      end else
      if YminPrev < Ymax-1 then
      begin
	YminPrev := iMax(YminPrev, Y);
	Ymax := iMin(Ymax, YminPrev+1);
      end;
      VerticalLine(X-1, YminPrev, YmaxPrev);
      if WXmax = WX2 then
      begin
	VerticalLine(X, Ymin, Ymax);
	exit;
      end;
      YminPrev := Ymin;
      YmaxPrev := Ymax;
    end;
  end;
end;

procedure TCustomZoomDisplay.PlotIntArray(
  const Samples: array of Integer;
  Offset: Longint;
  Channels: Integer;
  AChannel: Integer);
type
  TSample = Integer;
  PSample = ^TSample;
var
  AmpMin: Integer;
  AmpMax: Integer;
  X, Y, W, H: Integer;
  Ymin: Integer;
  Ymax: Integer;
  YminPrev: Integer;
  YmaxPrev: Integer;
  WH: Integer;
  WX: Integer;
  WY: Integer;
  WX1: Integer;
  WX2: Integer;
  WY2: Integer;
  WXmax: Integer;
  WXBufEnd: Integer;
  Sample: PSample;
  BufRem: Integer;
  Bits: PByteArray;
  ScanLineW: Integer;
  SampleCount: Integer;

  procedure VerticalLine(X, Ymin, Ymax: Integer);
  var
    Y: Integer;
    Mask: DWORD;
    ScanLineY: PByteArray;
  begin
    if X < 0 then Exit;
    Y := iMin(H-1, Ymin);
    if Y < 0 then Exit;
    Mask := $80 shr (X and 7);
    ScanLineY := PByteArray(Integer(Bits)+ScanLineW*(H-1-Y)); // PByteArray(Bitmap.ScanLine[Y]);
    for Y := Y downto iMax(0, Ymax) do
    begin
      ScanLineY^[X shr 3] := ScanLineY^[X shr 3] or Mask;
      Inc(PByte(ScanLineY), ScanLineW);
    end;
  end;

  procedure GetMinMax(Start, Stop: Integer);	assembler;
  {
  var
    i: Integer;
  begin
    for i := Start to Stop do
    begin
       WY := Sample^;
       if WY < AmpMin then AmpMin := WY;
       if WY > AmpMax then AmpMax := WY;
       Inc(Sample, Channels);
    end;
  end;
    // Min in ebx, Max in edx
  }
  asm
	mov	ecx, Stop
	sub	ecx, Start
	jl	@end
	inc	ecx
	push	ebx
	push	esi
	push	edi
	mov	esi, Sample
	mov	edi, Channels
	sal	edi, 2	// edi := Channels * SizeOf(Sample^)
	mov	ebx, AmpMin
	mov	edx, AmpMax
  @GetNext:
	mov	eax, [esi]
	add	esi, edi
	cmp	ebx, eax
	jng	@GetMax
	mov	ebx, eax
  @GetMax:
	cmp	edx, eax
	jng	@1
	dec	ecx
	jnz	@GetNext
	jmp	@exit
  @1:	mov	edx, eax
	dec     ecx
	jnz	@GetNext
  @exit:
	mov	Sample, esi
	mov	WY, eax
	mov	AmpMin, ebx
	mov	AmpMax, edx
	pop	edi
	pop	esi
	pop	ebx
  @end:
  end;

begin
  with Bitmap do
  begin
    if PixelFormat <> pf1bit then
    begin
      PixelFormat := pf1bit;
      SetBitmapColors(FBitmap, [Color, FPen.Color], 0);
    end;
    ScanLineW := BytesPerScanline(Width, 1, 32);
    Bits := ScanLine[Height-1];
    FillChar(Bits^, Height * ScanLineW, 0);
  end;
  WX1 := Window.Left;
  WX2 := Window.Right;
  Sample := PSample(@Samples);
  SampleCount := High(Samples)-Low(Samples)+1;
  if SampleCount < WX2-WX1+1 then Error(SPlotIntArray);
  BufRem := iMin(WX2-WX1+1, SampleCount-Offset);
  Inc(Sample, Offset * Channels + AChannel);
  WX := WX1;
  W := ClientWidth;
  if W >= WX2-WX1 then
  begin
    MoveTo(WX, Sample^);
    WXmax := WX + Pred(BufRem);
    for WX := Succ(WX) to WXmax do
    begin
      Inc(Sample, Channels);
      LineTo(WX, Sample^);
    end;
    Sample := PSample(@Samples);
    Inc(Sample, AChannel);
    BufRem := WX2-WX1+1-BufRem;
    for WX := Succ(WXmax) to WXmax + Pred(BufRem) do
    begin
      LineTo(WX, Sample^);
      Inc(Sample, Channels);
    end;
  end else
  begin
    H := ClientHeight;
    WY2 := Window.Top;
    WH := Window.Bottom-WY2;
    WY := Sample^;
    YMinPrev := IMulDiv(WY-WY2, H, WH);
    YmaxPrev := YminPrev;
    for X := 0 to W do
    begin
      Y := Trunc(((WY+Sample^)/2-WY2)*H/WH);
      AmpMin := High(TSample);
      AmpMax := Low(TSample);
      WXmax := iMin(WX2, WX1 + iMulDiv(X+1, WX2-WX1, W));
      WXBufEnd := WX+BufRem-1;
      if WXmax > WXBufEnd then
      begin
	GetMinMax(WX, WXBufEnd);
	WX := WXBufEnd+1;
	BufRem := iMin(WX2-WX+1, SampleCount);
	Sample := PSample(@Samples);
	Inc(Sample, AChannel);
      end;
      Dec(BufRem, WXmax-WX+1);
      GetMinMax(WX, WXmax);
      WX := WXmax+1;
      Ymin := IMulDiv(AmpMin-WY2, H, WH);
      Ymax := IMulDiv(AmpMax-WY2, H, WH);
      if Ymin < YmaxPrev-1 then
      begin
	Ymin := iMax(Ymin, Y);
	YmaxPrev := iMin(YmaxPrev, (Ymin+1));
      end else
      if YminPrev < Ymax-1 then
      begin
	YminPrev := iMax(YminPrev, Y);
	Ymax := iMin(Ymax, YminPrev+1);
      end;
      VerticalLine(X-1, YminPrev, YmaxPrev);
      if WXmax = WX2 then
      begin
	VerticalLine(X, Ymin, Ymax);
	exit;
      end;
      YminPrev := Ymin;
      YmaxPrev := Ymax;
    end;
  end;
end;

procedure TCustomZoomDisplay.Plot(
  P: Pointer;
  PType: Word;
  PInc: Integer;
  WorldX1, WorldX2: Integer;
  Func: TIntFunc);
///////////////// Under construction /////////////
var
  AmpMin: Integer;
  AmpMax: Integer;
  X, X1, H: Integer;
  Y: Single;
  Ymin: Integer;
  Ymax: Integer;
  YminPrev: Integer;
  YmaxPrev: Integer;
  WX, WY, WY2, WW, WH, WXmax: Integer;
  PB: PByte absolute P;
  PI: PSmallInt absolute P;
begin
  WorldX1 := iMax(WorldX1, Window.Left);
  WorldX2 := iMin(WorldX2, Window.Right);
  if WorldX1 = Window.Left
    then MoveTo(WorldX1, PI^)
    else LineTo(WorldX1, PI^);
  if ClientWidth >= Window.Right-Window.Left then
  begin
    Inc(PB, PInc);
    for WX := Succ(WorldX1) to WorldX2 do
    begin
      Inc(PB, PInc);
      LineTo(WX, PI^);
    end;
  end else
  begin
    H := ClientHeight;
    WY2 := Window.Top;
    WW := Window.Right-Window.Left;
    WH := Window.Bottom-WY2;
    X1 := DeviceX(WorldX1);
    WX := WorldX1;
    WY := PI^;
    YminPrev := IMulDiv(WY-WY2, H, WH);
    YmaxPrev := YminPrev;
    for X := X1 to DeviceX(WorldX2) do
    begin
      Y := ((WY+PI^)/2-WY2)*H/WH;
      AmpMin := High(Integer);
      AmpMax := Low(Integer);
      WXmax := iMin(Window.Right, Window.Left + iMulDiv(X+1, WW, ViewPort.W));
      for WX := WX to WXmax do
      begin
	 WY := PI^;
	 if WY < AmpMin then AmpMin := WY;
	 if WY > AmpMax then AmpMax := WY;
	 Inc(PB, PInc);
      end;
      WX := WXmax+1;
      Ymin := IMulDiv(AmpMin-WY2, H, WH);
      Ymax := IMulDiv(AmpMax-WY2, H, WH);
      if Ymin < YmaxPrev-1 then
      begin
	Ymin := iMax(Ymin, Trunc(Y));
	YmaxPrev := iMin(YmaxPrev, (Ymin+1));
      end else
      if YminPrev < Ymax-1 then
      begin
	YminPrev := iMax(YminPrev, Trunc(Y));
	Ymax := iMin(Ymax, YminPrev+1);
      end;
      DeviceMoveTo(X-1, YminPrev);
      DeviceLineTo(X-1, YmaxPrev-1);
      if WXmax = WorldX2 then
      begin
	DeviceMoveTo(X, Ymin);
	DeviceLineTo(X, Ymax-1);
	Exit;
      end;
      YminPrev := Ymin;
      YmaxPrev := Ymax;
    end;
  end;
end;

const
{$IFDEF Bilch}
  PalettePage = 'Homebrew';
{$ELSE}
  PalettePage = '3rdParty';
{$ENDIF}
procedure Register;
begin
  RegisterComponents(PalettePage, [TZoomDisplay]);
end;

initialization
  Screen.Cursors[crCrossHairs] := LoadCursor(HInstance, 'CROSSHAIRS');
  Screen.Cursors[crHand] := LoadCursor(HInstance, 'OPENHAND');
  Screen.Cursors[crGrip] := LoadCursor(HInstance, 'GRIP');
//  Screen.Cursors[crZoomIn] := LoadCursor(HInstance, 'ZOOMIN');
//  Screen.Cursors[crZoomOut] := LoadCursor(HInstance, 'ZOOMOUT');
end.

